# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

from math import ceil

from .base_method import BaseMethod
from ..utils import get_rng
from ..utils.common import safe_indexing


class RatioSplit(BaseMethod):
    """Splitting data into training, validation, and test sets based on provided sizes.
    Data is always shuffled before split.

    Parameters
    ----------
    data: array-like, required
        Raw preference data in the triplet format [(user_id, item_id, rating_value)].

    test_size: float, optional, default: 0.2
        The proportion of the test set, \
        if > 1 then it is treated as the size of the test set.

    val_size: float, optional, default: 0.0
        The proportion of the validation set, \
        if > 1 then it is treated as the size of the validation set.

    rating_threshold: float, optional, default: 1.0
        Threshold used to binarize rating values into positive or negative feedback for
        model evaluation using ranking metrics (rating metrics are not affected).

    seed: int, optional, default: None
        Random seed for reproducibility.

    exclude_unknowns: bool, optional, default: True
        If `True`, unknown users and items will be ignored during model evaluation.

    verbose: bool, optional, default: False
        Output running log.

    """

    def __init__(self, data, test_size=0.2, val_size=0.0, rating_threshold=1.0,
                 seed=None, exclude_unknowns=True, verbose=False, **kwargs):
        super().__init__(data=data, rating_threshold=rating_threshold, seed=seed,
                         exclude_unknowns=exclude_unknowns, verbose=verbose, **kwargs)

        self.train_size, self.val_size, self.test_size = self.validate_size(val_size, test_size, len(self._data))
        self._split()

    @staticmethod
    def validate_size(val_size, test_size, num_ratings):
        if val_size is None:
            val_size = 0.0
        elif val_size < 0:
            raise ValueError('val_size={} should be greater than zero'.format(val_size))
        elif val_size >= num_ratings:
            raise ValueError(
                'val_size={} should be less than the number of ratings {}'.format(val_size, num_ratings))

        if test_size is None:
            test_size = 0.0
        elif test_size < 0:
            raise ValueError('test_size={} should be greater than zero'.format(test_size))
        elif test_size >= num_ratings:
            raise ValueError(
                'test_size={} should be less than the number of ratings {}'.format(test_size, num_ratings))

        if val_size < 1:
            val_size = ceil(val_size * num_ratings)
        if test_size < 1:
            test_size = ceil(test_size * num_ratings)

        if val_size + test_size >= num_ratings:
            raise ValueError(
                'The sum of val_size and test_size ({}) should be smaller than the number of ratings {}'.format(
                    val_size + test_size, num_ratings))

        train_size = num_ratings - (val_size + test_size)

        return int(train_size), int(val_size), int(test_size)

    def _split(self):
        data_idx = self.rng.permutation(len(self._data))
        train_idx = data_idx[:self.train_size]
        test_idx = data_idx[-self.test_size:]
        val_idx = data_idx[self.train_size:-self.test_size]

        train_data = safe_indexing(self._data, train_idx)
        test_data = safe_indexing(self._data, test_idx)
        val_data = safe_indexing(self._data, val_idx) if len(val_idx) > 0 else None

        self.build(train_data=train_data, test_data=test_data, val_data=val_data)



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import numpy as np
from ..recommender import Recommender
from ...exception import ScoreException


class CVAECF(Recommender):
    """Conditional Variational Autoencoder for Collaborative Filtering.

    Parameters
    ----------
    z_dim: int, optional, default: 20
        The dimension of the stochastic user factors ``z'' representing the preference information.

    h_dim: int, optional, default: 20
        The dimension of the stochastic user factors ``h'' representing the auxiliary data.

    autoencoder_structure: list, default: [20]
        The number of neurons of encoder/decoder hidden layer for CVAE.
        For example, when autoencoder_structure = [20],
        the CVAE encoder structures will be [y_dim, 20, z_dim] and [x_dim, 20, h_dim],
        the decoder structure will be [z_dim + h_dim, 20, y_dim], where y and x are respectively the preference and \
        auxiliary data.

    act_fn: str, default: 'tanh'
        Name of the activation function used between hidden layers of the auto-encoder.
        Supported functions: ['sigmoid', 'tanh', 'elu', 'relu', 'relu6']

    likelihood: str, default: 'mult'
        Name of the likelihood function used for modeling user preferences.
        Supported choices:

        mult: Multinomial likelihood
        bern: Bernoulli likelihood
        gaus: Gaussian likelihood
        pois: Poisson likelihood

    n_epochs: int, optional, default: 100
        The number of epochs for SGD.

    batch_size: int, optional, default: 128
        The batch size.

    learning_rate: float, optional, default: 0.001
        The learning rate for Adam.

    beta: float, optional, default: 1.0
        The weight of the KL term KL(q(z|y)||p(z)) as in beta-VAE.

    alpha_1: float, optional, default: 1.0
        The weight of the KL term KL(q(h|x)||p(h|x)).

    alpha_2: float, optional, default: 1.0
        The weight of the KL term KL(q(h|x)||q(h|y)).

    name: string, optional, default: 'CVAECF'
        The name of the recommender model.

    trainable: boolean, optional, default: True
        When False, the model is not trained, and Cornac assumes that the model is already \
        pre-trained.

    verbose: boolean, optional, default: False
        When True, some running logs are displayed.

    seed: int, optional, default: None
        Random seed for parameters initialization.

    use_gpu: boolean, optional, default: False
        If True and your system supports CUDA then training is performed on GPUs.

    user auxiliary data : See "cornac/examples/cvaecf_filmtrust.py" for an example of how to use \
        cornac's graph modality to load and provide the ``user network'' for CVAECF.

    References
    ----------
    * Lee, Wonsung, Kyungwoo Song, and Il-Chul Moon. "Augmented variational autoencoders for collaborative filtering \
     with auxiliary information." Proceedings of ACM CIKM. 2017.
    """

    def __init__(
            self,
            name="CVAECF",
            z_dim=20,
            h_dim=20,
            autoencoder_structure=[20],
            act_fn="tanh",
            likelihood="mult",
            n_epochs=100,
            batch_size=128,
            learning_rate=0.001,
            beta=1.0,
            alpha_1=1.0,
            alpha_2=1.0,
            trainable=True,
            verbose=False,
            seed=None,
            use_gpu=False,
    ):
        Recommender.__init__(self, name=name, trainable=trainable, verbose=verbose)
        self.z_dim = z_dim
        self.h_dim = h_dim
        self.autoencoder_structure = autoencoder_structure
        self.act_fn = act_fn
        self.likelihood = likelihood
        self.batch_size = batch_size
        self.n_epochs = n_epochs
        self.learning_rate = learning_rate
        self.beta = beta
        self.alpha_1 = alpha_1
        self.alpha_2 = alpha_2
        self.seed = seed
        self.use_gpu = use_gpu

    def fit(self, train_set, val_set=None):
        """Fit the model to observations.

        Parameters
        ----------
        train_set: :obj:`cornac.data.Dataset`, required
            User-Item preference data as well as additional modalities.

        val_set: :obj:`cornac.data.Dataset`, optional, default: None
            User-Item preference data for model selection purposes (e.g., early stopping).

        Returns
        -------
        self : object
        """
        Recommender.fit(self, train_set, val_set)

        import torch
        from .cvaecf import CVAE, learn

        self.device = (
            torch.device("cuda:0")
            if (self.use_gpu and torch.cuda.is_available())
            else torch.device("cpu")
        )

        if self.trainable:
            if self.seed is not None:
                torch.manual_seed(self.seed)
                torch.cuda.manual_seed(self.seed)

            if not hasattr(self, "cvae"):
                n_items = train_set.matrix.shape[1]
                n_users = train_set.matrix.shape[0]
                self.cvae = CVAE(
                    self.z_dim,
                    self.h_dim,
                    [n_items] + self.autoencoder_structure,
                    [n_users] + self.autoencoder_structure,
                    self.act_fn,
                    self.likelihood,
                ).to(self.device)

            learn(
                self.cvae,
                self.train_set,
                n_epochs=self.n_epochs,
                batch_size=self.batch_size,
                learn_rate=self.learning_rate,
                beta=self.beta,
                alpha_1=self.alpha_1,
                alpha_2=self.alpha_2,
                verbose=self.verbose,
                device=self.device,
            )

        elif self.verbose:
            print("%s is trained already (trainable = False)" % (self.name))

        return self

    def score(self, user_idx, item_idx=None):
        """Predict the scores/ratings of a user for an item.

        Parameters
        ----------
        user_idx: int, required
            The index of the user for whom to perform score prediction.

        item_idx: int, optional, default: None
            The index of the item for which to perform score prediction.
            If None, scores for all known items will be returned.

        Returns
        -------
        res : A scalar or a Numpy array
            Relative scores that the user gives to the item or to all known items

        """
        import torch

        if item_idx is None:
            if self.train_set.is_unk_user(user_idx):
                raise ScoreException(
                    "Can't make score prediction for (user_id=%d)" % user_idx
                )

            y_u = self.train_set.matrix[user_idx].copy()
            y_u.data = np.ones(len(y_u.data))
            y_u = torch.tensor(y_u.A, dtype=torch.float32, device=self.device)
            z_u, _ = self.cvae.encode_qz(y_u)

            x_u = self.train_set.user_graph.matrix[user_idx].copy()
            x_u.data = np.ones(len(x_u.data))
            x_u = torch.tensor(x_u.A, dtype=torch.float32, device=self.device)
            h_u, _ = self.cvae.encode_qhx(x_u)

            known_item_scores = self.cvae.decode(z_u, h_u).data.cpu().numpy().flatten()

            return known_item_scores
        else:
            if self.train_set.is_unk_user(user_idx) or self.train_set.is_unk_item(
                    item_idx
            ):
                raise ScoreException(
                    "Can't make score prediction for (user_id=%d, item_id=%d)"
                    % (user_idx, item_idx)
                )

            y_u = self.train_set.matrix[user_idx].copy()
            y_u.data = np.ones(len(y_u.data))
            y_u = torch.tensor(y_u.A, dtype=torch.float32, device=self.device)
            z_u, _ = self.cvae.encode_qz(y_u)

            x_u = self.train_set.user_graph.matrix[user_idx].copy()
            x_u.data = np.ones(len(x_u.data))
            x_u = torch.tensor(x_u.A, dtype=torch.float32, device=self.device)
            h_u, _ = self.cvae.encode_qhx(x_u)

            user_pred = self.cvae.decode(z_u, h_u).data.cpu().numpy().flatten()[item_idx]

            return user_pred



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import numpy as np
from tqdm.auto import tqdm

from ..recommender import Recommender
from ...exception import CornacException
from ...exception import ScoreException
from ...utils import fast_dot
from ...utils.common import intersects
from ...utils import get_rng
from ...utils.init_utils import zeros, xavier_uniform


class VBPR(Recommender):
    """Visual Bayesian Personalized Ranking.

    Parameters
    ----------
    k: int, optional, default: 10
        The dimension of the gamma latent factors.

    k2: int, optional, default: 10
        The dimension of the theta latent factors.

    n_epochs: int, optional, default: 20
        Maximum number of epochs for SGD.

    batch_size: int, optional, default: 100
        The batch size for SGD.

    learning_rate: float, optional, default: 0.001
        The learning rate for SGD.

    lambda_w: float, optional, default: 0.01
        The regularization hyper-parameter for latent factor weights.

    lambda_b: float, optional, default: 0.01
        The regularization hyper-parameter for biases.

    lambda_e: float, optional, default: 0.0
        The regularization hyper-parameter for embedding matrix E and beta prime vector.

    use_gpu: boolean, optional, default: True
        Whether or not to use GPU to speed up training.

    trainable: boolean, optional, default: True
        When False, the model is not trained and Cornac assumes that the model already \
        pre-trained (U and V are not None).

    verbose: boolean, optional, default: True
        When True, running logs are displayed.

    init_params: dictionary, optional, default: None
        Initial parameters, e.g., init_params = {'Bi': beta_item, 'Gu': gamma_user,
        'Gi': gamma_item, 'Tu': theta_user, 'E': emb_matrix, 'Bp': beta_prime}

    seed: int, optional, default: None
        Random seed for weight initialization.

    References
    ----------
    * He, R., & McAuley, J. (2016). VBPR: Visual Bayesian Personalized Ranking from Implicit Feedback.
    """

    def __init__(
        self,
        name="VBPR",
        k=10,
        k2=10,
        n_epochs=50,
        batch_size=100,
        learning_rate=0.005,
        lambda_w=0.01,
        lambda_b=0.01,
        lambda_e=0.0,
        use_gpu=False,
        trainable=True,
        verbose=True,
        init_params=None,
        seed=None,
    ):
        super().__init__(name=name, trainable=trainable, verbose=verbose)
        self.k = k
        self.k2 = k2
        self.n_epochs = n_epochs
        self.batch_size = batch_size
        self.learning_rate = learning_rate
        self.lambda_w = lambda_w
        self.lambda_b = lambda_b
        self.lambda_e = lambda_e
        self.use_gpu = use_gpu
        self.seed = seed

        # Init params if provided
        self.init_params = {} if init_params is None else init_params
        self.beta_item = self.init_params.get("Bi", None)
        self.gamma_user = self.init_params.get("Gu", None)
        self.gamma_item = self.init_params.get("Gi", None)
        self.theta_user = self.init_params.get("Tu", None)
        self.emb_matrix = self.init_params.get("E", None)
        self.beta_prime = self.init_params.get("Bp", None)

    def _init(self, n_users, n_items, features):
        rng = get_rng(self.seed)

        self.beta_item = zeros(n_items) if self.beta_item is None else self.beta_item
        if self.gamma_user is None:
            self.gamma_user = xavier_uniform((n_users, self.k), rng)
        if self.gamma_item is None:
            self.gamma_item = xavier_uniform((n_items, self.k), rng)
        if self.theta_user is None:
            self.theta_user = xavier_uniform((n_users, self.k2), rng)
        if self.emb_matrix is None:
            self.emb_matrix = xavier_uniform((features.shape[1], self.k2), rng)
        if self.beta_prime is None:
            self.beta_prime = xavier_uniform((features.shape[1], 1), rng)

        # pre-computed for faster evaluation
        self.theta_item = np.matmul(features, self.emb_matrix)
        self.visual_bias = np.matmul(features, self.beta_prime).ravel()

    def fit(self, train_set, val_set=None):
        """Fit the model to observations.

        Parameters
        ----------
        train_set: :obj:`cornac.data.Dataset`, required
            User-Item preference data as well as additional modalities.

        val_set: :obj:`cornac.data.Dataset`, optional, default: None
            User-Item preference data for model selection purposes (e.g., early stopping).

        Returns
        -------
        self : object
        """
        Recommender.fit(self, train_set, val_set)

        if train_set.item_image is None:
            raise CornacException("item_image modality is required but None.")

        # Item visual feature from CNN
        train_features = train_set.item_image.features[: self.train_set.total_items]
        train_features = train_features.astype(np.float32)
        self._init(
            n_users=train_set.total_users,
            n_items=train_set.total_items,
            features=train_features,
        )

        if self.trainable:
            self._fit_torch(train_features)

        return self

    def _fit_torch(self, train_features):
        import torch

        def _l2_loss(*tensors):
            l2_loss = 0
            for tensor in tensors:
                l2_loss += tensor.pow(2).sum()
            return l2_loss / 2

        def _inner(a, b):
            return (a * b).sum(dim=1)

        dtype = torch.float
        device = (
            torch.device("cuda:0")
            if (self.use_gpu and torch.cuda.is_available())
            else torch.device("cpu")
        )

        F = torch.tensor(train_features, device=device, dtype=dtype)
        # Learned parameters
        Bi = torch.tensor(
            self.beta_item, device=device, dtype=dtype, requires_grad=True
        )
        Gu = torch.tensor(
            self.gamma_user, device=device, dtype=dtype, requires_grad=True
        )
        Gi = torch.tensor(
            self.gamma_item, device=device, dtype=dtype, requires_grad=True
        )
        Tu = torch.tensor(
            self.theta_user, device=device, dtype=dtype, requires_grad=True
        )
        E = torch.tensor(
            self.emb_matrix, device=device, dtype=dtype, requires_grad=True
        )
        Bp = torch.tensor(
            self.beta_prime, device=device, dtype=dtype, requires_grad=True
        )

        optimizer = torch.optim.Adam([Bi, Gu, Gi, Tu, E, Bp], lr=self.learning_rate)

        for epoch in range(1, self.n_epochs + 1):
            sum_loss = 0.0
            count = 0
            progress_bar = tqdm(
                total=self.train_set.num_batches(self.batch_size),
                desc="Epoch {}/{}".format(epoch, self.n_epochs),
                disable=not self.verbose,
            )
            for batch_u, batch_i, batch_j in self.train_set.uij_iter(
                self.batch_size, shuffle=True
            ):
                gamma_u = Gu[batch_u]
                theta_u = Tu[batch_u]

                beta_i = Bi[batch_i]
                beta_j = Bi[batch_j]
                gamma_i = Gi[batch_i]
                gamma_j = Gi[batch_j]
                feat_i = F[batch_i]
                feat_j = F[batch_j]

                gamma_diff = gamma_i - gamma_j
                feat_diff = feat_i - feat_j

                Xuij = (
                    beta_i
                    - beta_j
                    + _inner(gamma_u, gamma_diff)
                    + _inner(theta_u, feat_diff.mm(E))
                    + feat_diff.mm(Bp)
                )

                log_likelihood = torch.nn.functional.logsigmoid(Xuij).sum()

                reg = (
                    _l2_loss(gamma_u, gamma_i, gamma_j, theta_u) * self.lambda_w
                    + _l2_loss(beta_i) * self.lambda_b
                    + _l2_loss(beta_j) * self.lambda_b / 10
                    + _l2_loss(E, Bp) * self.lambda_e
                )

                loss = -log_likelihood + reg

                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

                sum_loss += loss.data.item()
                count += len(batch_u)
                if count % (self.batch_size * 10) == 0:
                    progress_bar.set_postfix(loss=(sum_loss / count))
                progress_bar.update(1)
            progress_bar.close()

        print("Optimization finished!")

        self.beta_item = Bi.data.cpu().numpy()
        self.gamma_user = Gu.data.cpu().numpy()
        self.gamma_item = Gi.data.cpu().numpy()
        self.theta_user = Tu.data.cpu().numpy()
        self.emb_matrix = E.data.cpu().numpy()
        self.beta_prime = Bp.data.cpu().numpy()
        # pre-computed for faster evaluation
        self.theta_item = F.mm(E).data.cpu().numpy()
        self.visual_bias = F.mm(Bp).data.cpu().numpy().ravel()

    def score(self, user_idx, item_idx=None):
        """Predict the scores/ratings of a user for an item.

        Parameters
        ----------
        user_idx: int, required
            The index of the user for whom to perform score prediction.

        item_idx: int, optional, default: None
            The index of the item for which to perform score prediction.
            If None, scores for all known items will be returned.

        Returns
        -------
        res : A scalar or a Numpy array
            Relative scores that the user gives to the item or to all known items

        """
        if item_idx is None:
            known_item_scores = np.add(self.beta_item, self.visual_bias)
            fast_dot(self.gamma_user[user_idx], self.gamma_item, known_item_scores)
            fast_dot(self.theta_user[user_idx], self.theta_item, known_item_scores)
            return known_item_scores
        else:
            item_score = np.add(self.beta_item[item_idx], self.visual_bias[item_idx])
            item_score += np.dot(self.gamma_item[item_idx], self.gamma_user[user_idx])
            item_score += np.dot(self.theta_item[item_idx], self.theta_user[user_idx])
            return item_score



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import numpy as np

from ...utils import get_rng

EPS = 1e-100


def _f_simplex(gamma, v, lambda_v, x):
    return 0.5 * lambda_v * np.dot((v - x).T, v - x) - np.sum(gamma * np.log(x))


def _df_simplex(gamma, v, lambda_v, x):
    return -lambda_v * (v - x) - np.sum(gamma * (1 / x), axis=0)


def _is_on_simplex(v, s):
    if v.sum() < s + 1e-10 and np.alltrue(v > 0):
        return True
    return False


def _simplex_project(v, s=1):
    assert s > 0, "Radius s must be strictly positive (%d <= 0)" % s
    n, = v.shape  # will raise ValueError if v is not 1-D
    # check if we are already on the simplex
    if _is_on_simplex(v, s):
        return v
    # get the array of cumulative sums of a sorted (decreasing) copy of v
    u = np.sort(v)[::-1]
    cssv = np.cumsum(u)
    # get the number of > 0 components of the optimal solution
    rho = np.nonzero(u * np.arange(1, n + 1) > (cssv - s))[0][-1]
    # compute the Lagrange multiplier associated to the simplex constraint
    theta = (cssv[rho] - s) / (rho + 1.0)
    # compute the projection by thresholding v using theta
    w = (v - theta).clip(min=0)
    return w


def _optimize_simplex(gamma, v, lambda_v, opt_x, s=1):
    opt_x_old = np.copy(opt_x)
    f_old = _f_simplex(gamma, v, lambda_v, opt_x)
    df = _df_simplex(gamma, v, lambda_v, opt_x)
    ab_sum = np.sum(np.absolute(df))
    if ab_sum > 1.0:
        df /= ab_sum
    opt_x -= df
    x_bar = _simplex_project(opt_x, s=s)
    x_bar -= opt_x_old
    r = 0.5 * np.dot(df, x_bar)
    beta = 0.5
    t = beta

    for iter in range(100):
        opt_x = np.copy(opt_x_old)
        opt_x += t * x_bar
        f_new = _f_simplex(gamma, v, lambda_v, opt_x)
        if f_new > f_old + r * t:
            t *= beta
        else:
            break

    if not _is_on_simplex(opt_x, s):
        print("Invalid values, outside simplex")

    return opt_x, f_new


class Model:
    def __init__(
        self,
        U,
        V,
        n_user,
        n_item,
        n_vocab,
        k=200,
        lambda_u=0.01,
        lambda_v=0.01,
        eta=0.01,
        a=1,
        b=0.01,
        max_iter=100,
        seed=None,
    ):
        self.k = k
        self.lambda_u = lambda_u
        self.lambda_v = lambda_v
        self.eta = eta
        self.a = a
        self.b = b
        self.max_iter = max_iter
        self.U = U
        self.V = V
        self.n_item = n_item
        self.n_user = n_user
        self.n_vocab = n_vocab
        self.seed = seed
        rng = get_rng(self.seed)

        # LDA variables
        self.theta = rng.random_sample([self.n_item, self.k])
        self.theta = self.theta / self.theta.sum(1)[:, np.newaxis]  # normalize
        self.beta = rng.random_sample([self.n_vocab, self.k])
        self.beta = self.beta / self.beta.sum(0)  # normalize
        self.phi_sum = np.zeros([self.n_vocab, self.k]) + self.eta

    def update_cf(self, user_data, item_data):
        R_user = user_data[1]
        R_item = item_data[1]

        cf_loss = 0.0

        # update users
        VV = self.b * (self.V.T.dot(self.V)) + self.lambda_u * np.eye(self.k)
        for i in range(self.n_user):
            idx_item = user_data[0][i]
            V_i = self.V[idx_item]
            R_i = R_user[i]
            A = VV + (self.a - self.b) * (V_i.T.dot(V_i))
            x = (self.a * V_i * (np.tile(R_i, (self.k, 1)).T)).sum(0)
            self.U[i] = np.linalg.solve(A, x)

            cf_loss += 0.5 * self.lambda_u * np.dot(self.U[i], self.U[i])

        # update items
        UU = self.b * (self.U.T.dot(self.U))
        for j in range(self.n_item):
            idx_user = item_data[0][j]
            U_j = self.U[idx_user]
            R_j = R_item[j]

            UU_j = UU + (self.a - self.b) * (U_j.T.dot(U_j))
            A = UU_j + self.lambda_v * np.eye(self.k)
            x = (self.a * U_j * (np.tile(R_j, (self.k, 1)).T)).sum(
                0
            ) + self.lambda_v * self.theta[j]
            self.V[j] = np.linalg.solve(A, x)

            cf_loss += 0.5 * self.a * np.square(R_j).sum()
            cf_loss -= self.a * np.sum((U_j.dot(self.V[j])) * R_j)
            cf_loss += 0.5 * np.dot(self.V[j].dot(UU_j), self.V[j])

            ep = self.V[j, :] - self.theta[j, :]
            cf_loss += 0.5 * self.lambda_v * np.sum(ep * ep)

        return cf_loss

    def update_beta(self):
        self.beta = self.phi_sum / self.phi_sum.sum(0)
        self.phi_sum = np.zeros([self.n_vocab, self.k]) + self.eta

    def update_theta(self, doc_ids, doc_cnt):
        loss = 0.0
        for vi in range(self.n_item):
            w = np.array(doc_ids[vi])
            word_beta = self.beta[w, :]
            phi = self.theta[vi, :] * word_beta + EPS  # W x K
            phi = phi / phi.sum(1)[:, np.newaxis]
            gamma = np.array(doc_cnt[vi])[:, np.newaxis] * phi

            self.theta[vi, :], lda_loss = _optimize_simplex(
                gamma=gamma,
                v=self.V[vi, :],
                opt_x=self.theta[vi, :],
                lambda_v=self.lambda_v,
                s=1,
            )
            loss += lda_loss

            self.phi_sum[w, :] += gamma

        return loss



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
This data is built based on the Amazon datasets provided by Julian McAuley @ http://jmcauley.ucsd.edu/data/amazon/.
We make sure all items having three types of auxiliary data: text, image, and context (items appearing together).
"""

from typing import List

import numpy as np

from ..utils import cache
from ..data import Reader
from ..data.reader import read_text


def load_feedback(reader: Reader = None) -> List:
    """Load the user-item ratings, scale: [1,5]

    Parameters
    ----------
    reader: `obj:cornac.data.Reader`, default: None
        Reader object used to read the data.

    Returns
    -------
    data: array-like
        Data in the form of a list of tuples (user, item, rating).
    """
    fpath = cache(url='https://static.preferred.ai/cornac/datasets/amazon_clothing/rating.zip',
                  unzip=True, relative_path='amazon_clothing/rating.txt')
    reader = Reader() if reader is None else reader
    return reader.read(fpath, sep='\t')


def load_text():
    """Load the item text descriptions

    Returns
    -------
    texts: List
        List of text documents, one per item.

    ids: List
        List of item ids aligned with indices in `texts`.
    """
    fpath = cache(url='https://static.preferred.ai/cornac/datasets/amazon_clothing/text.zip',
                  unzip=True, relative_path='amazon_clothing/text.txt')
    texts, ids = read_text(fpath, sep='::')
    return texts, ids


def load_visual_feature():
    """Load item visual features (extracted from pre-trained CNN)

    Returns
    -------
    features: numpy.ndarray
        Feature matrix with shape (n, 4096) with n is the number of items.

    item_ids: List
        List of item ids aligned with indices in `features`.
    """
    features = np.load(cache(url='https://static.preferred.ai/cornac/datasets/amazon_clothing/image.zip',
                             unzip=True, relative_path='amazon_clothing/image_features.npy'))
    item_ids = read_text(cache(url='https://static.preferred.ai/cornac/datasets/amazon_clothing/item_ids.zip',
                               unzip=True, relative_path='amazon_clothing/item_ids.txt'))
    return features, item_ids


def load_graph(reader: Reader = None) -> List:
    """Load the item-item interactions (symmetric network), built from the Amazon Also-Viewed information

    Parameters
    ----------
    reader: `obj:cornac.data.Reader`, default: None
        Reader object used to read the data.

    Returns
    -------
    data: array-like
        Data in the form of a list of tuples (item, item, 1).
    """
    fpath = cache(url='https://static.preferred.ai/cornac/datasets/amazon_clothing/context.zip',
                  unzip=True, relative_path='amazon_clothing/context.txt')
    reader = Reader() if reader is None else reader
    return reader.read(fpath, fmt='UI', sep='\t')



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import os
import copy
import inspect
import pickle
from glob import glob
from datetime import datetime

import numpy as np

from ..exception import ScoreException
from ..utils.common import intersects, clip


class Recommender:
    """Generic class for a recommender model. All recommendation models should inherit from this class 
    
    Parameters
    ----------------
    name: str, required
        The name of the recommender model

    trainable: boolean, optional, default: True
        When False, the model is not trainable

    """

    def __init__(self, name, trainable=True, verbose=False):
        self.name = name
        self.trainable = trainable
        self.verbose = verbose
        self.train_set = None
        self.val_set = None
        # attributes to be ignored when being saved
        self.ignored_attrs = ["train_set", "val_set"]

    def reset_info(self):
        self.best_value = -np.Inf
        self.best_epoch = 0
        self.current_epoch = 0
        self.stopped_epoch = 0
        self.wait = 0

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        for k, v in self.__dict__.items():
            if k in self.ignored_attrs:
                continue
            setattr(result, k, copy.deepcopy(v))
        return result

    @classmethod
    def _get_init_params(cls):
        """Get initial parameters from the model constructor"""
        init = getattr(cls.__init__, "deprecated_original", cls.__init__)
        if init is object.__init__:
            return []

        init_signature = inspect.signature(init)
        parameters = [p for p in init_signature.parameters.values() if p.name != "self"]

        return sorted([p.name for p in parameters])

    def clone(self, new_params=None):
        """Clone an instance of the model object.

        Parameters
        ----------
        new_params: dict, optional, default: None
            New parameters for the cloned instance.

        Returns
        -------
        object: :obj:`cornac.models.Recommender`
        """
        new_params = {} if new_params is None else new_params
        init_params = {}
        for name in self._get_init_params():
            init_params[name] = new_params.get(name, copy.deepcopy(getattr(self, name)))

        return self.__class__(**init_params)

    def save(self, save_dir=None):
        """Save a recommender model to the filesystem.

        Parameters
        ----------
        save_dir: str, default: None
            Path to a directory for the model to be stored.

        Returns
        -------
        model_file : str
            Path to the model file stored on the filesystem.
        """
        if save_dir is None:
            return

        model_dir = os.path.join(save_dir, self.name)
        os.makedirs(model_dir, exist_ok=True)
        timestamp = datetime.now().strftime("%Y-%m-%d_%H-%M-%S-%f")
        model_file = os.path.join(model_dir, "{}.pkl".format(timestamp))

        saved_model = copy.deepcopy(self)

        pickle.dump(
            saved_model, open(model_file, "wb"), protocol=pickle.HIGHEST_PROTOCOL
        )

        if self.verbose:
            print("{} model is saved to {}".format(self.name, model_file))

        return model_file

    @staticmethod
    def load(model_path, trainable=False):
        """Load a recommender model from the filesystem.

        Parameters
        ----------
        model_path: str, required
            Path to a file or directory where the model is stored. If a directory is
            provided, the latest model will be loaded.

        trainable: boolean, optional, default: False
            Set it to True if you would like to finetune the model. By default, 
            the model parameters are assumed to be fixed after being loaded.
        
        Returns
        -------
        self : object
        """
        if os.path.isdir(model_path):
            model_file = sorted(glob("{}/*.pkl".format(model_path)))[-1]
        else:
            model_file = model_path

        model = pickle.load(open(model_file, "rb"))
        model.trainable = trainable
        model.load_from = model_file  # for further loading

        return model

    def fit(self, train_set, val_set=None):
        """Fit the model to observations.

        Parameters
        ----------
        train_set: :obj:`cornac.data.Dataset`, required
            User-Item preference data as well as additional modalities.

        val_set: :obj:`cornac.data.Dataset`, optional, default: None
            User-Item preference data for model selection purposes (e.g., early stopping).

        Returns
        -------
        self : object
        """
        self.reset_info()
        self.train_set = train_set.reset()
        self.val_set = None if val_set is None else val_set.reset()
        return self

    def score(self, user_idx, item_idx=None):
        """Predict the scores/ratings of a user for an item.

        Parameters
        ----------
        user_idx: int, required
            The index of the user for whom to perform score prediction.
            
        item_idx: int, optional, default: None
            The index of the item for which to perform score prediction.
            If None, scores for all known items will be returned.

        Returns
        -------
        res : A scalar or a Numpy array
            Relative scores that the user gives to the item or to all known items

        """
        raise NotImplementedError("The algorithm is not able to make score prediction!")

    def default_score(self):
        """Overwrite this function if your algorithm has special treatment for cold-start problem

        """
        return self.train_set.global_mean

    def rate(self, user_idx, item_idx, clipping=True):
        """Give a rating score between pair of user and item

        Parameters
        ----------
        user_idx: int, required
            The index of the user for whom to perform item raking.

        item_idx: int, required
            The index of the item to be rated by the user.

        clipping: bool, default: True
            Whether to clip the predicted rating value.

        Returns
        -------
        A scalar
            A rating score of the user for the item
        """
        try:
            rating_pred = self.score(user_idx, item_idx)
        except ScoreException:
            rating_pred = self.default_score()

        if clipping:
            rating_pred = clip(
                values=rating_pred,
                lower_bound=self.train_set.min_rating,
                upper_bound=self.train_set.max_rating,
            )

        return rating_pred

    def rank(self, user_idx, item_indices=None):
        """Rank all test items for a given user.

        Parameters
        ----------
        user_idx: int, required
            The index of the user for whom to perform item raking.

        item_indices: 1d array, optional, default: None
            A list of candidate item indices to be ranked by the user.
            If `None`, list of ranked known item indices and their scores will be returned.

        Returns
        -------
        (item_rank, item_scores): tuple
            `item_rank` contains item indices being ranked by their scores.
            `item_scores` contains scores of items corresponding to their indices in the `item_indices` input.
        """
        # obtain item scores from the model
        try:
            known_item_scores = self.score(user_idx)
        except ScoreException:
            known_item_scores = (
                np.ones(self.train_set.total_items) * self.default_score()
            )

        # check if the returned scores also cover unknown items
        # if not, all unknown items will be given the MIN score
        if len(known_item_scores) == self.train_set.total_items:
            all_item_scores = known_item_scores
        else:
            all_item_scores = np.ones(self.train_set.total_items) * np.min(
                known_item_scores
            )
            all_item_scores[: self.train_set.num_items] = known_item_scores

        # rank items based on their scores
        if item_indices is None:
            item_scores = all_item_scores[: self.train_set.num_items]
            item_rank = item_scores.argsort()[::-1]
        else:
            item_scores = all_item_scores[item_indices]
            item_rank = np.array(item_indices)[item_scores.argsort()[::-1]]

        return item_rank, item_scores

    def monitor_value(self):
        """Calculating monitored value used for early stopping on validation set (`val_set`).
        This function will be called by `early_stop()` function.
        Note: `val_set` could be `None` thus it needs to be checked before usage.

        Returns
        -------
        :raise NotImplementedError
        """
        raise NotImplementedError()

    def early_stop(self, min_delta=0.0, patience=0):
        """Check if training should be stopped when validation loss has stopped improving.

        Parameters
        ----------
        min_delta: float, optional, default: 0.
            The minimum increase in monitored value on validation set to be considered as improvement,
            i.e. an increment of less than `min_delta` will count as no improvement.

        patience: int, optional, default: 0
            Number of epochs with no improvement after which training should be stopped.

        Returns
        -------
        res : bool
            Return `True` if model training should be stopped (no improvement on validation set),
            otherwise return `False`.
        """
        self.current_epoch += 1
        current_value = self.monitor_value()
        if current_value is None:
            return False

        if np.greater_equal(current_value - self.best_value, min_delta):
            self.best_value = current_value
            self.best_epoch = self.current_epoch
            self.wait = 0
        else:
            self.wait += 1
            if self.wait >= patience:
                self.stopped_epoch = self.current_epoch

        if self.stopped_epoch > 0:
            print("Early stopping:")
            print(
                "- best epoch = {}, stopped epoch = {}".format(
                    self.best_epoch, self.stopped_epoch
                )
            )
            print(
                "- best monitored value = {:.6f} (delta = {:.6f})".format(
                    self.best_value, current_value - self.best_value
                )
            )
            return True
        return False



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import os
import time
import math

import numpy as np
from tqdm.auto import trange

from ..recommender import Recommender
from ...exception import ScoreException
from ...utils import get_rng
from ...utils.init_utils import xavier_uniform


class ConvMF(Recommender):
    """
    Parameters
    ----------
    k: int, optional, default: 50
        The dimension of the user and item latent factors.

    n_epochs: int, optional, default: 50
        Maximum number of epochs for training.

    cnn_epochs: int, optional, default: 5
        Number of epochs for optimizing the CNN for each overall training epoch.
    
    cnn_bs: int, optional, default: 128
        Batch size for optimizing CNN.
        
    cnn_lr: float, optional, default: 0.001
        Learning rate for optimizing CNN.

    lambda_u: float, optional, default: 1.0
        The regularization hyper-parameter for user latent factor.

    lambda_v: float, optional, default: 100.0
        The regularization hyper-parameter for item latent factor.

    emb_dim: int, optional, default: 200
        The embedding size of each word. One word corresponds with [1 x emb_dim] vector in the embedding space

    max_len: int, optional, default 300
        The maximum length of item's document

    filter_sizes: list, optional, default: [3, 4, 5]
        The length of filters in convolutional layer

    num_filters: int, optional, default: 100
        The number of filters in convolutional layer
        
    hidden_dim: int, optional, default: 200
        The dimension of hidden layer after the pooling of all convolutional layers

    dropout_rate: float, optional, default: 0.2
        Dropout rate while training CNN

    give_item_weight: boolean, optional, default: True
        When True, each item will be weighted base on the number of user who have rated this item

    init_params: dict, optional, default: {'U':None, 'V':None, 'W': None}
        Initial U and V matrix and initial weight for embedding layer W

    trainable: boolean, optional, default: True
        When False, the model is not trained and Cornac assumes that the model already \
        pre-trained (U and V are not None).

    References
    ----------
    * Donghyun Kim1, Chanyoung Park1. ConvMF: Convolutional Matrix Factorization for Document Context-Aware Recommendation. \
    In :10th ACM Conference on Recommender Systems Pages 233-240
    """

    def __init__(
        self,
        name="ConvMF",
        k=50,
        n_epochs=50,
        cnn_epochs=5,
        cnn_bs=128,
        cnn_lr=0.001,
        lambda_u=1,
        lambda_v=100,
        emb_dim=200,
        max_len=300,
        filter_sizes=[3, 4, 5],
        num_filters=100,
        hidden_dim=200,
        dropout_rate=0.2,
        give_item_weight=True,
        trainable=True,
        verbose=False,
        init_params=None,
        seed=None,
    ):
        super().__init__(name=name, trainable=trainable, verbose=verbose)
        self.give_item_weight = give_item_weight
        self.n_epochs = n_epochs
        self.cnn_bs = cnn_bs
        self.cnn_lr = cnn_lr
        self.lambda_u = lambda_u
        self.lambda_v = lambda_v
        self.k = k
        self.dropout_rate = dropout_rate
        self.emb_dim = emb_dim
        self.max_len = max_len
        self.filter_sizes = filter_sizes
        self.num_filters = num_filters
        self.hidden_dim = hidden_dim
        self.name = name
        self.verbose = verbose
        self.cnn_epochs = cnn_epochs
        self.seed = seed

        # Init params if provided
        self.init_params = {} if init_params is None else init_params
        self.U = self.init_params.get("U", None)
        self.V = self.init_params.get("V", None)
        self.W = self.init_params.get("W", None)

    def _init(self):
        rng = get_rng(self.seed)
        n_users, n_items = self.train_set.num_users, self.train_set.num_items
        vocab_size = self.train_set.item_text.vocab.size

        if self.U is None:
            self.U = xavier_uniform((n_users, self.k), rng)
        if self.V is None:
            self.V = xavier_uniform((n_items, self.k), rng)
        if self.W is None:
            self.W = xavier_uniform((vocab_size, self.emb_dim), rng)

    def fit(self, train_set, val_set=None):
        """Fit the model to observations.

        Parameters
        ----------
        train_set: :obj:`cornac.data.Dataset`, required
            User-Item preference data as well as additional modalities.

        val_set: :obj:`cornac.data.Dataset`, optional, default: None
            User-Item preference data for model selection purposes (e.g., early stopping).

        Returns
        -------
        self : object
        """
        Recommender.fit(self, train_set, val_set)

        self._init()

        if self.trainable:
            self._fit_convmf()

        return self

    @staticmethod
    def _build_data(csr_mat):
        data = []
        index_list = []
        rating_list = []
        for i in range(csr_mat.shape[0]):
            j, k = csr_mat.indptr[i], csr_mat.indptr[i + 1]
            index_list.append(csr_mat.indices[j:k])
            rating_list.append(csr_mat.data[j:k])
        data.append(index_list)
        data.append(rating_list)
        return data

    def _fit_convmf(self):
        user_data = self._build_data(self.train_set.matrix)
        item_data = self._build_data(self.train_set.matrix.T.tocsr())

        n_user = len(user_data[0])
        n_item = len(item_data[0])

        # R_user and R_item contain rating values
        R_user = user_data[1]
        R_item = item_data[1]

        if self.give_item_weight:
            item_weight = np.array([math.sqrt(len(i)) for i in R_item], dtype=float)
            item_weight = (float(n_item) / item_weight.sum()) * item_weight
        else:
            item_weight = np.ones(n_item, dtype=float)

        # Initialize cnn module
        import tensorflow.compat.v1 as tf
        from .convmf import CNN_module

        tf.disable_eager_execution()

        # less verbose TF
        os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
        tf.logging.set_verbosity(tf.logging.ERROR)

        tf.set_random_seed(self.seed)
        cnn_module = CNN_module(
            output_dimension=self.k,
            dropout_rate=self.dropout_rate,
            emb_dim=self.emb_dim,
            max_len=self.max_len,
            filter_sizes=self.filter_sizes,
            num_filters=self.num_filters,
            hidden_dim=self.hidden_dim,
            seed=self.seed,
            init_W=self.W,
            learning_rate=self.cnn_lr,
        )

        config = tf.ConfigProto()
        config.gpu_options.allow_growth = True
        sess = tf.Session(config=config)

        sess.run(tf.global_variables_initializer())  # init variable

        document = self.train_set.item_text.batch_seq(
            np.arange(n_item), max_length=self.max_len
        )

        feed_dict = {cnn_module.model_input: document}
        theta = sess.run(cnn_module.model_output, feed_dict=feed_dict)

        endure = 3
        converge_threshold = 0.01
        history = 1e-50
        loss = 0

        for epoch in range(1, self.n_epochs + 1):
            if self.verbose:
                print("Epoch: {}/{}".format(epoch, self.n_epochs))

            tic = time.time()

            user_loss = 0.0
            for i in range(n_user):
                idx_item = user_data[0][i]
                V_i = self.V[idx_item]
                R_i = R_user[i]

                A = self.lambda_u * np.eye(self.k) + V_i.T.dot(V_i)
                B = (V_i * (np.tile(R_i, (self.k, 1)).T)).sum(0)
                self.U[i] = np.linalg.solve(A, B)

                user_loss += self.lambda_u * np.dot(self.U[i], self.U[i])

            item_loss = 0.0
            for j in range(n_item):
                idx_user = item_data[0][j]
                U_j = self.U[idx_user]
                R_j = R_item[j]

                A = self.lambda_v * item_weight[j] * np.eye(self.k) + U_j.T.dot(U_j)
                B = (U_j * (np.tile(R_j, (self.k, 1)).T)).sum(
                    0
                ) + self.lambda_v * item_weight[j] * theta[j]
                self.V[j] = np.linalg.solve(A, B)

                item_loss += np.square(R_j - U_j.dot(self.V[j])).sum()

            loop = trange(
                self.cnn_epochs, desc="Optimizing CNN", disable=not self.verbose
            )
            for _ in loop:
                for batch_ids in self.train_set.item_iter(
                    batch_size=self.cnn_bs, shuffle=True
                ):
                    batch_seq = self.train_set.item_text.batch_seq(
                        batch_ids, max_length=self.max_len
                    )
                    feed_dict = {
                        cnn_module.model_input: batch_seq,
                        cnn_module.v: self.V[batch_ids],
                        cnn_module.sample_weight: item_weight[batch_ids],
                    }

                    sess.run([cnn_module.optimizer], feed_dict=feed_dict)

            feed_dict = {
                cnn_module.model_input: document,
                cnn_module.v: self.V,
                cnn_module.sample_weight: item_weight,
            }
            theta, cnn_loss = sess.run(
                [cnn_module.model_output, cnn_module.weighted_loss], feed_dict=feed_dict
            )

            loss = 0.5 * (user_loss + item_loss + self.lambda_v * cnn_loss)

            toc = time.time()
            elapsed = toc - tic
            converge = abs((loss - history) / history)

            if self.verbose:
                print(
                    "Loss: %.5f Elapsed: %.4fs Converge: %.6f "
                    % (loss, elapsed, converge)
                )

            history = loss
            if converge < converge_threshold:
                endure -= 1
                if endure == 0:
                    break

        tf.reset_default_graph()

    def score(self, user_idx, item_idx=None):
        """Predict the scores/ratings of a user for an item.

        Parameters
        ----------
        user_idx: int, required
            The index of the user for whom to perform score prediction.

        item_idx: int, optional, default: None
            The index of the item for which to perform score prediction.
            If None, scores for all known items will be returned.

        Returns
        -------
        res : A scalar or a Numpy array
            Relative scores that the user gives to the item or to all known items

        """
        if item_idx is None:
            if self.train_set.is_unk_user(user_idx):
                raise ScoreException(
                    "Can't make score prediction for (user_id=%d)" % user_idx
                )

            known_item_scores = self.V.dot(self.U[user_idx, :])
            return known_item_scores
        else:
            if self.train_set.is_unk_user(user_idx) or self.train_set.is_unk_item(
                item_idx
            ):
                raise ScoreException(
                    "Can't make score prediction for (user_id=%d, item_id=%d)"
                    % (user_idx, item_idx)
                )

            user_pred = self.V[item_idx, :].dot(self.U[user_idx, :])

            return user_pred



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import numpy as np

from ..recommender import Recommender
from ...utils.common import scale
from ...exception import ScoreException


class BiVAECF(Recommender):
    """Bilateral Variational AutoEncoder for Collaborative Filtering.

    Parameters
    ----------
    k: int, optional, default: 10
        The dimension of the stochastic user ``theta'' and item ``beta'' factors.

    encoder_structure: list, default: [20]
        The number of neurons per layer of the user and item encoders for BiVAE.
        For example, encoder_structure = [20], the user (item) encoder structure will be [num_items, 20, k] ([num_users, 20, k]).

    act_fn: str, default: 'tanh'
        Name of the activation function used between hidden layers of the auto-encoder.
        Supported functions: ['sigmoid', 'tanh', 'elu', 'relu', 'relu6']

    likelihood: str, default: 'pois'
        The likelihood function used for modeling the observations.
        Supported choices:

        bern: Bernoulli likelihood
        gaus: Gaussian likelihood
        pois: Poisson likelihood

    n_epochs: int, optional, default: 100
        The number of epochs for SGD.

    batch_size: int, optional, default: 100
        The batch size.

    learning_rate: float, optional, default: 0.001
        The learning rate for Adam.

    beta_kl: float, optional, default: 1.0
        The weight of the KL terms as in beta-VAE.

    cap_priors: dict, optional, default: {"user":False, "item":False}
        When {"user":True, "item":True}, CAP priors are used (see BiVAE paper for details),\
        otherwise the standard Normal is used as a Prior over the user and item latent variables.

    name: string, optional, default: 'BiVAECF'
        The name of the recommender model.

    trainable: boolean, optional, default: True
        When False, the model is not trained and Cornac assumes that the model is already \
        pre-trained.

    verbose: boolean, optional, default: False
        When True, some running logs are displayed.

    seed: int, optional, default: None
        Random seed for parameters initialization.

    use_gpu: boolean, optional, default: True
        If True and your system supports CUDA then training is performed on GPUs.

    References
    ----------
    * Quoc-Tuan Truong, Aghiles Salah, Hady W. Lauw. " Bilateral Variational Autoencoder for Collaborative Filtering."
    ACM International Conference on Web Search and Data Mining (WSDM). 2021.
    """

    def __init__(
        self,
        name="BiVAECF",
        k=10,
        encoder_structure=[20],
        act_fn="tanh",
        likelihood="pois",
        n_epochs=100,
        batch_size=100,
        learning_rate=0.001,
        beta_kl=1.0,
        cap_priors={"user": False, "item": False},
        trainable=True,
        verbose=False,
        seed=None,
        use_gpu=True,
    ):
        Recommender.__init__(self, name=name, trainable=trainable, verbose=verbose)
        self.k = k
        self.encoder_structure = encoder_structure
        self.act_fn = act_fn
        self.likelihood = likelihood
        self.batch_size = batch_size
        self.n_epochs = n_epochs
        self.learning_rate = learning_rate
        self.beta_kl = beta_kl
        self.cap_priors = cap_priors
        self.seed = seed
        self.use_gpu = use_gpu

    def fit(self, train_set, val_set=None):
        """Fit the model to observations.

        Parameters
        ----------
        train_set: :obj:`cornac.data.Dataset`, required
            User-Item preference data as well as additional modalities.

        val_set: :obj:`cornac.data.Dataset`, optional, default: None
            User-Item preference data for model selection purposes (e.g., early stopping).

        Returns
        -------
        self : object
        """
        Recommender.fit(self, train_set, val_set)

        import torch
        from .bivae import BiVAE, learn

        self.device = (
            torch.device("cuda:0")
            if (self.use_gpu and torch.cuda.is_available())
            else torch.device("cpu")
        )

        if self.trainable:
            feature_dim = {"user": None, "item": None}
            if self.cap_priors.get("user", False):
                if train_set.user_feature is None:
                    raise ValueError(
                        "CAP priors for users is set to True but no user features are provided"
                    )
                else:
                    feature_dim["user"] = train_set.user_feature.feature_dim

            if self.cap_priors.get("item", False):
                if train_set.item_feature is None:
                    raise ValueError(
                        "CAP priors for items is set to True but no item features are provided"
                    )
                else:
                    feature_dim["item"] = train_set.item_feature.feature_dim

            if self.seed is not None:
                torch.manual_seed(self.seed)
                torch.cuda.manual_seed(self.seed)

            if not hasattr(self, "bivaecf"):
                num_items = train_set.matrix.shape[1]
                num_users = train_set.matrix.shape[0]
                self.bivae = BiVAE(
                    k=self.k,
                    user_encoder_structure=[num_items] + self.encoder_structure,
                    item_encoder_structure=[num_users] + self.encoder_structure,
                    act_fn=self.act_fn,
                    likelihood=self.likelihood,
                    cap_priors=self.cap_priors,
                    feature_dim=feature_dim,
                    batch_size=self.batch_size,
                ).to(self.device)

            learn(
                self.bivae,
                self.train_set,
                n_epochs=self.n_epochs,
                batch_size=self.batch_size,
                learn_rate=self.learning_rate,
                beta_kl=self.beta_kl,
                verbose=self.verbose,
                device=self.device,
            )

        elif self.verbose:
            print("%s is trained already (trainable = False)" % (self.name))

        return self

    def score(self, user_idx, item_idx=None):
        """Predict the scores/ratings of a user for an item.

        Parameters
        ----------
        user_idx: int, required
            The index of the user for whom to perform score prediction.

        item_idx: int, optional, default: None
            The index of the item for which to perform score prediction.
            If None, scores for all known items will be returned.

        Returns
        -------
        res : A scalar or a Numpy array
            Relative scores that the user gives to the item or to all known items

        """

        if item_idx is None:
            if self.train_set.is_unk_user(user_idx):
                raise ScoreException(
                    "Can't make score prediction for (user_id=%d)" % user_idx
                )

            theta_u = self.bivae.mu_theta[user_idx].view(1, -1)
            beta = self.bivae.mu_beta
            known_item_scores = (
                self.bivae.decode_user(theta_u, beta).cpu().numpy().ravel()
            )

            return known_item_scores
        else:
            if self.train_set.is_unk_user(user_idx) or self.train_set.is_unk_item(
                item_idx
            ):
                raise ScoreException(
                    "Can't make score prediction for (user_id=%d, item_id=%d)"
                    % (user_idx, item_idx)
                )

            theta_u = self.bivae.mu_theta[user_idx].view(1, -1)
            beta_i = self.bivae.mu_beta[item_idx].view(1, -1)
            pred = self.bivae.decode_user(theta_u, beta_i).cpu().numpy().ravel()

            pred = scale(
                pred, self.train_set.min_rating, self.train_set.max_rating, 0.0, 1.0
            )

            return pred



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from tqdm.auto import trange

from ...utils import estimate_batches

torch.set_default_dtype(torch.float32)

EPS = 1e-10

ACT = {
    "sigmoid": nn.Sigmoid(),
    "tanh": nn.Tanh(),
    "elu": nn.ELU(),
    "relu": nn.ReLU(),
    "relu6": nn.ReLU6(),
}


class VAE(nn.Module):
    def __init__(self, z_dim, ae_structure, act_fn, likelihood):
        super(VAE, self).__init__()

        self.likelihood = likelihood
        self.act_fn = ACT.get(act_fn, None)
        if self.act_fn is None:
            raise ValueError("Supported act_fn: {}".format(ACT.keys()))

        # Encoder
        self.encoder = nn.Sequential()
        for i in range(len(ae_structure) - 1):
            self.encoder.add_module(
                "fc{}".format(i), nn.Linear(ae_structure[i], ae_structure[i + 1])
            )
            self.encoder.add_module("act{}".format(i), self.act_fn)
        self.enc_mu = nn.Linear(ae_structure[-1], z_dim)  # mu
        self.enc_logvar = nn.Linear(ae_structure[-1], z_dim)  # logvar

        # Decoder
        ae_structure = [z_dim] + ae_structure[::-1]
        self.decoder = nn.Sequential()
        for i in range(len(ae_structure) - 1):
            self.decoder.add_module(
                "fc{}".format(i), nn.Linear(ae_structure[i], ae_structure[i + 1])
            )
            if i != len(ae_structure) - 2:
                self.decoder.add_module("act{}".format(i), self.act_fn)

    def encode(self, x):
        h = self.encoder(x)
        return self.enc_mu(h), self.enc_logvar(h)

    def decode(self, z):
        h = self.decoder(z)
        if self.likelihood == "mult":
            return torch.softmax(h, dim=1)
        else:
            return torch.sigmoid(h)

    def reparameterize(self, mu, logvar):
        std = torch.exp(0.5 * logvar)
        eps = torch.randn_like(mu)
        return mu + eps * std

    def forward(self, x):
        mu, logvar = self.encode(x)
        z = self.reparameterize(mu, logvar)
        return self.decode(z), mu, logvar

    def loss(self, x, x_, mu, logvar, beta):
        # Likelihood
        ll_choices = {
            "mult": x * torch.log(x_ + EPS),
            "bern": x * torch.log(x_ + EPS) + (1 - x) * torch.log(1 - x_ + EPS),
            "gaus": -(x - x_) ** 2,
            "pois": x * torch.log(x_ + EPS) - x_,
        }

        ll = ll_choices.get(self.likelihood, None)
        if ll is None:
            raise ValueError("Supported likelihoods: {}".format(ll_choices.keys()))

        ll = torch.sum(ll, dim=1)

        # KL term
        std = torch.exp(0.5 * logvar)
        kld = -0.5 * (1 + 2.0 * torch.log(std) - mu.pow(2) - std.pow(2))
        kld = torch.sum(kld, dim=1)

        return torch.mean(beta * kld - ll)


def learn(
    vae,
    train_set,
    n_epochs,
    batch_size,
    learn_rate,
    beta,
    verbose,
    device=torch.device("cpu"),
):
    optimizer = torch.optim.Adam(params=vae.parameters(), lr=learn_rate)
    num_steps = estimate_batches(train_set.num_users, batch_size)

    progress_bar = trange(1, n_epochs + 1, disable=not verbose)
    for _ in progress_bar:
        sum_loss = 0.0
        count = 0
        for batch_id, u_ids in enumerate(
            train_set.user_iter(batch_size, shuffle=False)
        ):
            u_batch = train_set.matrix[u_ids, :]
            u_batch.data = np.ones(len(u_batch.data))  # Binarize data
            u_batch = u_batch.A
            u_batch = torch.tensor(u_batch, dtype=torch.float32, device=device)

            # Reconstructed batch
            u_batch_, mu, logvar = vae(u_batch)

            loss = vae.loss(u_batch, u_batch_, mu, logvar, beta)
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            sum_loss += loss.data.item()
            count += len(u_batch)

            if batch_id % 10 == 0:
                progress_bar.set_postfix(loss=(sum_loss / count))

    return vae



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import numpy as np
import torch


def online_ibpr(
    train_set,
    k,
    lamda=0.005,
    n_epochs=150,
    learning_rate=0.001,
    batch_size=100,
    init_params=None,
):
    X = train_set.matrix
    # change the data to original user Id item Id and rating format
    X = X.tocoo()  # convert sparse matrix to COOrdiante format
    triplets = np.ndarray(shape=(len(X.data), 3), dtype=float)
    triplets[:, 0] = X.row
    triplets[:, 1] = X.col
    triplets[:, 2] = X.data

    # Initial user factors
    if init_params["U"] is None:
        U = torch.randn(X.shape[0], k, requires_grad=True)
    else:
        U = torch.from_numpy(init_params["U"])
        U.requires_grad = True

    # Initial item factors
    if init_params["V"] is None:
        V = torch.randn(X.shape[1], k, requires_grad=True)
    else:
        V = torch.from_numpy(init_params["V"])
        V.requires_grad = True

    optimizer = torch.optim.Adam([U], lr=learning_rate)
    for epoch in range(n_epochs):
        regU = U[triplets[:, 0], :]
        regI = V[triplets[:, 1], :]
        regJ = V[triplets[:, 2], :]

        regU_unq = U[np.unique(triplets[:, 0]), :]
        regI_unq = V[np.unique(triplets[:, 1:]), :]

        regU_norm = regU / regU.norm(dim=1)[:, None]
        regI_norm = regI / regI.norm(dim=1)[:, None]
        regJ_norm = regJ / regJ.norm(dim=1)[:, None]

        Scorei = torch.acos(
            torch.clamp(torch.sum(regU_norm * regI_norm, dim=1), -1 + 1e-7, 1 - 1e-7)
        )
        Scorej = torch.acos(
            torch.clamp(torch.sum(regU_norm * regJ_norm, dim=1), -1 + 1e-7, 1 - 1e-7)
        )

        loss = (
            lamda * (regU_unq.norm().pow(2) + regI_unq.norm().pow(2))
            - torch.log(torch.sigmoid(Scorej - Scorei)).sum()
        )
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()
        print("epoch:", epoch, "loss:", loss)

    # since the user's preference is defined by the angular distance, we can normalize the user/item vectors without changing the ranking
    # U = torch.nn.functional.normalize(U, p = 2, dim=1)
    # V = torch.nn.functional.normalize(V, p = 2, dim=1)
    U = U.data.cpu().numpy()
    V = V.data.cpu().numpy()

    res = {"U": U, "V": V}

    return res



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================


import numpy as np
import torch
from tqdm.auto import tqdm


def ibpr(
    train_set,
    k,
    lamda=0.001,
    n_epochs=150,
    learning_rate=0.05,
    batch_size=100,
    init_params=None,
    verbose=False,
):
    X = train_set.csr_matrix

    # Initial user factors
    if init_params["U"] is None:
        U = torch.randn(X.shape[0], k, requires_grad=True)
    else:
        U = torch.from_numpy(init_params["U"])
        U.requires_grad = True

    # Initial item factors
    if init_params["V"] is None:
        V = torch.randn(X.shape[1], k, requires_grad=True)
    else:
        V = torch.from_numpy(init_params["V"])
        V.requires_grad = True
        
    optimizer = torch.optim.Adam([U, V], lr=learning_rate)

    for epoch in range(1, n_epochs + 1):
        sum_loss = 0.0
        count = 0
        progress_bar = tqdm(
            total=train_set.num_batches(batch_size),
            desc="Epoch {}/{}".format(epoch, n_epochs),
            disable=not verbose,
        )

        for batch_u, batch_i, batch_j in train_set.uij_iter(batch_size, shuffle=True):
            regU = U[batch_u, :]
            regI = V[batch_i, :]
            regJ = V[batch_j, :]

            regU_unq = U[np.unique(batch_u), :]
            regI_unq = V[np.union1d(batch_i, batch_j), :]

            regU_norm = regU / regU.norm(dim=1)[:, None]
            regI_norm = regI / regI.norm(dim=1)[:, None]
            regJ_norm = regJ / regJ.norm(dim=1)[:, None]

            Scorei = torch.acos(
                torch.clamp(
                    torch.sum(regU_norm * regI_norm, dim=1), -1 + 1e-7, 1 - 1e-7
                )
            )
            Scorej = torch.acos(
                torch.clamp(
                    torch.sum(regU_norm * regJ_norm, dim=1), -1 + 1e-7, 1 - 1e-7
                )
            )

            loss = (
                lamda * (regU_unq.norm().pow(2) + regI_unq.norm().pow(2))
                - torch.log(torch.sigmoid(Scorej - Scorei)).sum()
            )
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            sum_loss += loss.data.item()
            count += len(batch_u)
            if count % (batch_size * 10) == 0:
                progress_bar.set_postfix(loss=(sum_loss / count))
            progress_bar.update(1)

        progress_bar.close()

    # since the user's preference is defined by the angular distance,
    # we can normalize the user/item vectors without changing the ranking
    U = torch.nn.functional.normalize(U, p=2, dim=1)
    V = torch.nn.functional.normalize(V, p=2, dim=1)
    U = U.data.cpu().numpy()
    V = V.data.cpu().numpy()

    res = {"U": U, "V": V}

    return res



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import numbers

import numpy as np
import scipy.sparse as sp

from .fast_sparse_funcs import (
    inplace_csr_row_normalize_l1,
    inplace_csr_row_normalize_l2
)

FLOAT_DTYPES = (np.float64, np.float32, np.float16)


def sigmoid(x):
    """Sigmoid function"""
    return 1. / (1. + np.exp(-x))


def scale(values, target_min, target_max, source_min=None, source_max=None):
    """Scale the value of a numpy array "values"
    from source_min, source_max into a range [target_min, target_max]

    Parameters
    ----------
    values : Numpy array, required
        Values to be scaled.

    target_min : scalar, required
        Target minimum value.

    target_max : scalar, required
        Target maximum value.

    source_min : scalar, required, default: None
        Source minimum value if desired. If None, it will be the minimum of values.

    source_max : scalar, required, default: None
        Source minimum value if desired. If None, it will be the maximum of values.

    Returns
    -------
    res: Numpy array
        Output values mapped into range [target_min, target_max]
    """
    if source_min is None:
        source_min = np.min(values)
    if source_max is None:
        source_max = np.max(values)
    if source_min == source_max:  # improve this scenario
        source_min = 0.0

    values = (values - source_min) / (source_max - source_min)
    values = values * (target_max - target_min) + target_min
    return values


def clip(values, lower_bound, upper_bound):
    """Perform clipping to enforce values to lie
    in a specific range [lower_bound, upper_bound]

    Parameters
    ----------
    values : Numpy array, required
        Values to be clipped.

    lower_bound : scalar, required
        Lower bound of the output.

    upper_bound : scalar, required
        Upper bound of the output.

    Returns
    -------
    res: Numpy array
        Clipped values in range [lower_bound, upper_bound]
    """
    values = np.where(values > upper_bound, upper_bound, values)
    values = np.where(values < lower_bound, lower_bound, values)

    return values


def intersects(x, y, assume_unique=False):
    """Return the intersection of given two arrays
    """
    mask = np.in1d(x, y, assume_unique=assume_unique)
    x_intersects_y = x[mask]

    return x_intersects_y


def excepts(x, y, assume_unique=False):
    """Removing elements in array y from array x
    """
    mask = np.in1d(x, y, assume_unique=assume_unique, invert=True)
    x_excepts_y = x[mask]

    return x_excepts_y


def safe_indexing(X, indices):
    """Return items or rows from X using indices.
    Allows simple indexing of lists or arrays.
    Borrowed from https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/utils/__init__.py

    Parameters
    ----------
    X : array-like, sparse-matrix, list, pandas.DataFrame, pandas.Series.
        Data from which to sample rows or items.
    indices : array-like of int
        Indices according to which X will be subsampled.

    Returns
    -------
    subset
        Subset of X on first axis
    """
    if hasattr(X, "shape"):
        if hasattr(X, 'take') and (hasattr(indices, 'dtype') and
                                   indices.dtype.kind == 'i'):
            # This is often substantially faster than X[indices]
            return X.take(indices, axis=0)
        else:
            return X[indices]
    else:
        return [X[idx] for idx in indices]


def validate_format(input_format, valid_formats):
    """Check the input format is in list of valid formats
    :raise ValueError if not supported
    """
    if not input_format in valid_formats:
        raise ValueError('{} data format is not in valid formats ({})'.format(input_format, valid_formats))

    return input_format


def estimate_batches(input_size, batch_size):
    """
    Estimate number of batches give `input_size` and `batch_size`
    """
    return int(np.ceil(input_size / batch_size))


def get_rng(seed):
    '''Return a RandomState of Numpy.
    If seed is None, use RandomState singleton from numpy.
    If seed is an integer, create a RandomState from that seed.
    If seed is already a RandomState, just return it.
    '''
    if seed is None:
        return np.random.mtrand._rand
    if isinstance(seed, (numbers.Integral, np.integer)):
        return np.random.RandomState(seed)
    if isinstance(seed, np.random.RandomState):
        return seed
    raise ValueError('{} can not be used to create a numpy.random.RandomState'.format(seed))


def normalize(X, norm='l2', axis=1, copy=True):
    """Scale input vectors individually to unit norm (vector length).

    Parameters
    ----------
    X : {array-like, sparse matrix}, shape [n_samples, n_features]
        The data to normalize, element by element.
        scipy.sparse matrices should be in CSR format to avoid an
        un-necessary copy.

    norm : 'l1', 'l2', or 'max', optional ('l2' by default)
        The norm to use to normalize each non zero sample (or each non-zero
        feature if axis is 0).

    axis : 0 or 1, optional (1 by default)
        axis used to normalize the data along. If 1, independently normalize
        each sample, otherwise (if 0) normalize each feature.

    copy : boolean, optional, default True
        set to False to perform inplace row normalization and avoid a
        copy (if the input is already a numpy array or a scipy.sparse
        CSR matrix and if axis is 1).
        
    Reference
    ---------
    https://github.com/scikit-learn/scikit-learn/blob/1495f69242646d239d89a5713982946b8ffcf9d9/sklearn/preprocessing/data.py#L1553

    """
    if norm not in ('l1', 'l2', 'max'):
        raise ValueError("'%s' is not a supported norm" % norm)

    if len(X.shape) != 2:
        raise ValueError("input X must be 2D but shape={}".format(X.shape))

    X_out = X.copy() if copy else X
    X_out = X_out if X_out.dtype in FLOAT_DTYPES else X_out.astype(np.float64)

    if axis == 0:
        X_out = X_out.T

    if sp.issparse(X_out):
        X_out = X_out.tocsr()

        if norm == 'l1':
            inplace_csr_row_normalize_l1(X_out)
        elif norm == 'l2':
            inplace_csr_row_normalize_l2(X_out)
        elif norm == 'max':
            norms = X_out.max(axis=1).A
            norms_elementwise = norms.repeat(np.diff(X_out.indptr))
            mask = norms_elementwise != 0
            X_out.data[mask] /= norms_elementwise[mask]
    else:
        if norm == 'l1':
            norms = np.abs(X_out).sum(axis=1)
        elif norm == 'l2':
            norms = np.sqrt((X_out ** 2).sum(axis=1))
        elif norm == 'max':
            norms = np.max(X_out, axis=1)
        norms[norms == 0] = 1.
        X_out /= norms.reshape(-1, 1)

    if axis == 0:
        X_out = X_out.T

    return X_out



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import numpy as np
import scipy.sparse as sp

from cornac.models.hpf import hpf
from ..recommender import Recommender
from ...exception import ScoreException


class HPF(Recommender):
    """Hierarchical Poisson Factorization.

    Parameters
    ----------
    k: int, optional, default: 5
        The dimension of the latent factors.

    max_iter: int, optional, default: 100
        Maximum number of iterations.

    name: string, optional, default: 'HPF'
        The name of the recommender model.

    trainable: boolean, optional, default: True
        When False, the model is not trained and Cornac assumes that the model is already \
        pre-trained (Theta and Beta are not None). 
        
    verbose: boolean, optional, default: False
        When True, some running logs are displayed.
        
    hierarchical: boolean, optional, default: True
        When False, PF is used instead of HPF.
        
    seed: int, optional, default: None
        Random seed for parameters initialization.

    init_params: dict, optional, default: None
        Initial parameters of the model.
        
        Theta: ndarray, shape (n_users, k)
            The expected user latent factors.
        
        Beta: ndarray, shape (n_items, k)
            The expected item latent factors.

        G_s: ndarray, shape (n_users, k)
            This represents "shape" parameters of Gamma distribution over Theta.

        G_r: ndarray, shape (n_users, k)
            This represents "rate" parameters of Gamma distribution over Theta.
            
        L_s: ndarray, shape (n_items, k)
            This represents "shape" parameters of Gamma distribution over Beta.
        
        L_r: ndarray, shape (n_items, k)
            This represents "rate" parameters of Gamma distribution over Beta.
        
    References
    ----------
    * Gopalan, Prem, Jake M. Hofman, and David M. Blei. Scalable Recommendation with \
    Hierarchical Poisson Factorization. In UAI, pp. 326-335. 2015.
    """

    def __init__(
        self,
        k=5,
        max_iter=100,
        name="HPF",
        trainable=True,
        verbose=False,
        hierarchical=True,
        seed=None,
        init_params=None,
    ):
        Recommender.__init__(self, name=name, trainable=trainable, verbose=verbose)
        self.k = k
        self.max_iter = max_iter

        self.ll = np.full(max_iter, 0)
        self.etp_r = np.full(max_iter, 0)
        self.etp_c = np.full(max_iter, 0)
        self.eps = 0.000000001
        self.hierarchical = hierarchical
        self.seed = seed

        # Init params if provided
        self.init_params = {} if init_params is None else init_params
        self.Theta = self.init_params.get("Theta", None)  # matrix of user factors
        self.Beta = self.init_params.get("Beta", None)  # matrix of item factors
        self.Gs = self.init_params.get("G_s", None)
        self.Gr = self.init_params.get("G_r", None)
        self.Ls = self.init_params.get("L_s", None)
        self.Lr = self.init_params.get("L_r", None)

    def fit(self, train_set, val_set=None):
        """Fit the model to observations.

        Parameters
        ----------
        train_set: :obj:`cornac.data.Dataset`, required
            User-Item preference data as well as additional modalities.

        val_set: :obj:`cornac.data.Dataset`, optional, default: None
            User-Item preference data for model selection purposes (e.g., early stopping).

        Returns
        -------
        self : object
        """
        Recommender.fit(self, train_set, val_set)

        if self.trainable:
            # use pre-trained params if exists, otherwise from constructor
            init_params = {
                "G_s": self.Gs,
                "G_r": self.Gr,
                "L_s": self.Ls,
                "L_r": self.Lr,
            }

            X = sp.csc_matrix(self.train_set.matrix)
            # recover the striplet sparse format from csc sparse matrix X (needed to feed c++)
            (rid, cid, val) = sp.find(X)
            val = np.array(val, dtype="float32")
            rid = np.array(rid, dtype="int32")
            cid = np.array(cid, dtype="int32")
            tX = np.concatenate(
                (np.concatenate(([rid], [cid]), axis=0).T, val.reshape((len(val), 1))),
                axis=1,
            )
            del rid, cid, val

            if self.hierarchical:
                res = hpf.hpf(
                    tX, X.shape[0], X.shape[1], self.k, self.max_iter, self.seed, init_params
                )
            else:
                res = hpf.pf(
                    tX, X.shape[0], X.shape[1], self.k, self.max_iter, self.seed, init_params
                )
            self.Theta = np.asarray(res["Z"])
            self.Beta = np.asarray(res["W"])

            # overwrite init_params for future fine-tuning
            self.Gs = np.asarray(res["G_s"])
            self.Gr = np.asarray(res["G_r"])
            self.Ls = np.asarray(res["L_s"])
            self.Lr = np.asarray(res["L_r"])

        elif self.verbose:
            print("%s is trained already (trainable = False)" % (self.name))

        return self

    def score(self, user_idx, item_idx=None):
        """Predict the scores/ratings of a user for an item.

        Parameters
        ----------
        user_idx: int, required
            The index of the user for whom to perform score prediction.

        item_idx: int, optional, default: None
            The index of the item for which to perform score prediction.
            If None, scores for all known items will be returned.

        Returns
        -------
        res : A scalar or a Numpy array
            Relative scores that the user gives to the item or to all known items

        """
        if item_idx is None:
            if self.train_set.is_unk_user(user_idx):
                u_representation = np.ones(self.k)
            else:
                u_representation = self.Theta[user_idx, :]

            known_item_scores = self.Beta.dot(u_representation)
            known_item_scores = np.array(known_item_scores, dtype="float64").flatten()
            return known_item_scores
        else:
            if self.train_set.is_unk_user(user_idx) or self.train_set.is_unk_item(
                item_idx
            ):
                raise ScoreException(
                    "Can't make score prediction for (user_id=%d, item_id=%d)"
                    % (user_idx, item_idx)
                )

            user_pred = self.Beta[item_idx, :].dot(self.Theta[user_idx, :])
            user_pred = np.array(user_pred, dtype="float64").flatten()[0]

            return user_pred



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import numpy as np
import scipy

from ..recommender import Recommender
from ...utils.common import sigmoid
from ...utils.common import scale
from ...exception import ScoreException


class SoRec(Recommender):
    """Social recommendation using Probabilistic Matrix Factorization.

    Parameters
    ----------
    k: int, optional, default: 5
        The dimension of the latent factors.

    max_iter: int, optional, default: 100
        Maximum number of iterations or the number of epochs for SGD.

    learning_rate: float, optional, default: 0.001
        The learning rate for SGD_RMSProp.

    gamma: float, optional, default: 0.9
        The weight for previous/current gradient in RMSProp.

    lambda_c: float, optional, default: 10
        The parameter balancing the information from the user-item rating matrix and the user social network.

    lambda_reg: float, optional, default: 0.001
        The regularization parameter.
        
    weight_link: boolean, optional, default: True
        When true the social network links are weighted according to eq. (4) in the original paper.

    name: string, optional, default: 'SoRec'
        The name of the recommender model.

    trainable: boolean, optional, default: True
        When False, the model is not trained and Cornac assumes that the model already
        pre-trained (U, V and Z are not None).

    verbose: boolean, optional, default: False
        When True, some running logs are displayed.

    init_params: dictionary, optional, default: None
        List of initial parameters, e.g., init_params = {'U':U, 'V':V, 'Z':Z}.

        U: a ndarray of shape (n_users, k)
            Containing the user latent factors.
            
        V: a ndarray of shape (n_items, k)
            Containing the item latent factors.
            
        Z: a ndarray of shape (n_users, k)
            Containing the social network latent factors.

    seed: int, optional, default: None
        Random seed for parameters initialization.

    References
    ----------
    * H. Ma, H. Yang, M. R. Lyu, and I. King. SoRec:Social recommendation using probabilistic matrix factorization. \
     CIKM ’08, pages 931–940, Napa Valley, USA, 2008.

    """

    def __init__(
        self,
        name="SoRec",
        k=5,
        max_iter=100,
        learning_rate=0.001,
        lambda_c=10,
        lambda_reg=0.001,
        gamma=0.9,
        weight_link=True,
        trainable=True,
        verbose=False,
        init_params=None,
        seed=None,
    ):
        Recommender.__init__(self, name=name, trainable=trainable, verbose=verbose)
        self.k = k
        self.max_iter = max_iter
        self.learning_rate = learning_rate
        self.lambda_c = lambda_c
        self.lambda_reg = lambda_reg
        self.gamma = gamma
        self.weight_link = weight_link

        self.ll = np.full(max_iter, 0)
        self.eps = 0.000000001
        self.seed = seed

        # Init params if provided
        self.init_params = {} if init_params is None else init_params
        self.U = self.init_params.get("U", None)  # matrix of user factors
        self.V = self.init_params.get("V", None)  # matrix of item factors
        self.Z = self.init_params.get("Z", None)  # matrix of social network factors

        if self.U is not None and self.U.shape[1] != self.k:
            raise ValueError("initial parameters U dimension error")

        if self.V is not None and self.V.shape[1] != self.k:
            raise ValueError("initial parameters V dimension error")

        if self.Z is not None and self.Z.shape[1] != self.k:
            raise ValueError("initial parameters Z dimension error")

    def fit(self, train_set, val_set=None):
        """Fit the model to observations.

        Parameters
        ----------
        train_set: :obj:`cornac.data.Dataset`, required
            User-Item preference data as well as additional modalities.

        val_set: :obj:`cornac.data.Dataset`, optional, default: None
            User-Item preference data for model selection purposes (e.g., early stopping).

        Returns
        -------
        self : object
        """
        Recommender.fit(self, train_set, val_set)

        import math
        from cornac.models.sorec import sorec

        if self.trainable:
            # user-item interactions
            (rat_uid, rat_iid, rat_val) = train_set.uir_tuple

            # user social network
            map_uid = train_set.user_indices
            (net_uid, net_jid, net_val) = train_set.user_graph.get_train_triplet(
                map_uid, map_uid
            )

            if self.weight_link:
                degree = train_set.user_graph.get_node_degree(map_uid, map_uid)
                weighted_net_val = []
                for u, j, val in zip(net_uid, net_jid, net_val):
                    u_out = degree[int(u)][1]
                    j_in = degree[int(j)][0]
                    val_weighted = math.sqrt(j_in / (j_in + u_out)) * val
                    weighted_net_val.append(val_weighted)
                net_val = weighted_net_val

            if [self.train_set.min_rating, self.train_set.max_rating] != [0, 1]:
                if self.train_set.min_rating == self.train_set.max_rating:
                    rat_val = scale(rat_val, 0.0, 1.0, 0.0, self.train_set.max_rating)
                else:
                    rat_val = scale(
                        rat_val,
                        0.0,
                        1.0,
                        self.train_set.min_rating,
                        self.train_set.max_rating,
                    )

            rat_val = np.array(rat_val, dtype="float32")
            rat_uid = np.array(rat_uid, dtype="int32")
            rat_iid = np.array(rat_iid, dtype="int32")

            net_val = np.array(net_val, dtype="float32")
            net_uid = np.array(net_uid, dtype="int32")
            net_jid = np.array(net_jid, dtype="int32")

            if self.verbose:
                print("Learning...")

            res = sorec.sorec(
                rat_uid,
                rat_iid,
                rat_val,
                net_uid,
                net_jid,
                net_val,
                k=self.k,
                n_users=train_set.num_users,
                n_items=train_set.num_items,
                n_ratings=len(rat_val),
                n_edges=len(net_val),
                n_epochs=self.max_iter,
                lambda_c=self.lambda_c,
                lambda_reg=self.lambda_reg,
                learning_rate=self.learning_rate,
                gamma=self.gamma,
                init_params={"U": self.U, "V": self.V, "Z": self.Z},
                verbose=self.verbose,
                seed=self.seed,
            )

            self.U = np.asarray(res["U"])
            self.V = np.asarray(res["V"])
            self.Z = np.asarray(res["Z"])

            if self.verbose:
                print("Learning completed")
                
        elif self.verbose:
            print("%s is trained already (trainable = False)" % self.name)

        return self

    def score(self, user_idx, item_idx=None):
        """Predict the scores/ratings of a user for an item.
        Parameters
        ----------
        user_idx: int, required
            The index of the user for whom to perform score prediction.
        item_idx: int, optional, default: None
            The index of the item for which to perform score prediction.
            If None, scores for all known items will be returned.
        Returns
        -------
        res : A scalar or a Numpy array
            Relative scores that the user gives to the item or to all known items
        """
        if item_idx is None:
            if self.train_set.is_unk_user(user_idx):
                raise ScoreException(
                    "Can't make score prediction for (user_id=%d)" % user_idx
                )

            known_item_scores = self.V.dot(self.U[user_idx, :])
            return known_item_scores
        else:
            if self.train_set.is_unk_user(user_idx) or self.train_set.is_unk_item(
                item_idx
            ):
                raise ScoreException(
                    "Can't make score prediction for (user_id=%d, item_id=%d)"
                    % (user_idx, item_idx)
                )

            user_pred = self.V[item_idx, :].dot(self.U[user_idx, :])
            user_pred = sigmoid(user_pred)
            if self.train_set.min_rating == self.train_set.max_rating:
                user_pred = scale(user_pred, 0.0, self.train_set.max_rating, 0.0, 1.0)
            else:
                user_pred = scale(
                    user_pred,
                    self.train_set.min_rating,
                    self.train_set.max_rating,
                    0.0,
                    1.0,
                )

            return user_pred



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import tensorflow.compat.v1 as tf

from ...utils import get_rng
from ...utils.init_utils import xavier_uniform

act_functions = {
    "sigmoid": tf.nn.sigmoid,
    "tanh": tf.nn.tanh,
    "elu": tf.nn.elu,
    "relu": tf.nn.relu,
    "relu6": tf.nn.relu6,
    "leaky_relu": tf.nn.leaky_relu,
    "identity": tf.identity,
}


# Stacked Denoising Autoencoder
def sdae(X_corrupted, layers, dropout_rate=0.0, act_fn="relu", seed=None, name="SDAE"):
    fn = act_functions.get(act_fn, None)
    if fn is None:
        raise ValueError(
            "Invalid type of activation function {}\n"
            "Supported functions: {}".format(act_fn, act_functions.keys())
        )

    # Weight initialization
    L = len(layers)
    rng = get_rng(seed)
    weights, biases = [], []
    with tf.variable_scope(name):
        for i in range(L - 1):
            w = xavier_uniform((layers[i], layers[i + 1]), random_state=rng)
            weights.append(
                tf.get_variable(
                    name="W_{}".format(i), dtype=tf.float32, initializer=tf.constant(w)
                )
            )
            biases.append(
                tf.get_variable(
                    name="b_{}".format(i),
                    dtype=tf.float32,
                    shape=(layers[i + 1]),
                    initializer=tf.zeros_initializer(),
                )
            )

    # Construct the auto-encoder
    h_value = X_corrupted
    for i in range(L - 1):
        # The encoder
        if i <= int(L / 2) - 1:
            h_value = fn(tf.add(tf.matmul(h_value, weights[i]), biases[i]))
        # The decoder
        elif i > int(L / 2) - 1:
            h_value = fn(tf.add(tf.matmul(h_value, weights[i]), biases[i]))
        # The dropout for all the layers except the final one
        if i < (L - 2):
            h_value = tf.nn.dropout(h_value, rate=dropout_rate)
        # The encoder output value
        if i == int(L / 2) - 1:
            encoded_value = h_value

    sdae_value = h_value

    # L2 loss
    loss_w = tf.constant(0.0)
    for i in range(L - 1):
        loss_w = tf.add(loss_w, tf.nn.l2_loss(weights[i]) + tf.nn.l2_loss(biases[i]))

    return sdae_value, encoded_value, loss_w


class Model:
    def __init__(
        self,
        n_users,
        n_items,
        n_vocab,
        k,
        layers,
        lambda_u,
        lambda_v,
        lambda_n,
        lambda_w,
        lr,
        dropout_rate,
        U,
        V,
        act_fn,
        seed,
    ):
        self.n_vocab = n_vocab
        self.n_users = n_users
        self.n_items = n_items
        self.lambda_u = lambda_u
        self.lambda_v = lambda_v
        self.lambda_n = lambda_n
        self.lambda_w = lambda_w
        self.layers = layers
        self.lr = lr  # learning rate
        self.k = k  # latent dimension
        self.dropout_rate = dropout_rate
        self.U_init = tf.constant(U)
        self.V_init = tf.constant(V)
        self.act_fn = act_fn
        self.seed = seed

        self._build_graph()

    def _build_graph(self):
        self.text_mask = tf.placeholder(
            dtype=tf.float32, shape=[None, self.n_vocab], name="mask_input"
        )
        self.text_input = tf.placeholder(
            dtype=tf.float32, shape=[None, self.n_vocab], name="text_input"
        )
        self.ratings = tf.placeholder(
            dtype=tf.float32, shape=[self.n_users, None], name="rating_input"
        )
        self.C = tf.placeholder(
            dtype=tf.float32, shape=[self.n_users, None], name="C_input"
        )

        with tf.variable_scope("CDL_Variable"):
            self.U = tf.get_variable(
                name="U", dtype=tf.float32, initializer=self.U_init
            )
            self.V = tf.get_variable(
                name="V", dtype=tf.float32, initializer=self.V_init
            )

        self.item_ids = tf.placeholder(dtype=tf.int32)
        real_batch_size = tf.cast(tf.shape(self.text_input)[0], tf.int32)
        V_batch = tf.reshape(
            tf.gather(self.V, self.item_ids), shape=[real_batch_size, self.k]
        )

        corrupted_text = tf.multiply(self.text_input, self.text_mask)
        sdae_value, encoded_value, loss_w = sdae(
            X_corrupted=corrupted_text,
            layers=self.layers,
            dropout_rate=self.dropout_rate,
            act_fn=self.act_fn,
            seed=self.seed,
            name="SDAE_Variable",
        )

        loss_1 = self.lambda_u * tf.nn.l2_loss(self.U) + self.lambda_w * loss_w
        loss_2 = self.lambda_v * tf.nn.l2_loss(V_batch - encoded_value)
        loss_3 = self.lambda_n * tf.nn.l2_loss(sdae_value - self.text_input)

        predictions = tf.matmul(self.U, V_batch, transpose_b=True)
        squared_error = tf.square(self.ratings - predictions)
        loss_4 = tf.reduce_sum(tf.multiply(self.C, squared_error))

        self.loss = loss_1 + loss_2 + loss_3 + loss_4

        # Generate optimizer
        optimizer1 = tf.train.AdamOptimizer(self.lr)
        optimizer2 = tf.train.AdamOptimizer(self.lr)

        train_var_list1, train_var_list2 = [], []

        for var in tf.trainable_variables():
            if "CDL_Variable" in var.name:
                train_var_list1.append(var)
            elif "SDAE_Variable" in var.name:
                train_var_list2.append(var)

        gvs = optimizer1.compute_gradients(self.loss, var_list=train_var_list1)
        capped_gvs = [(tf.clip_by_value(grad, -5.0, 5.0), var) for grad, var in gvs]
        self.opt1 = optimizer1.apply_gradients(capped_gvs)

        gvs = optimizer2.compute_gradients(self.loss, var_list=train_var_list2)
        capped_gvs = [(tf.clip_by_value(grad, -5.0, 5.0), var) for grad, var in gvs]
        self.opt2 = optimizer2.apply_gradients(capped_gvs)



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import numpy as np
from tqdm.auto import tqdm
from scipy.special import expit

from ..recommender import Recommender
from ...exception import CornacException
from ...exception import ScoreException
from ...utils import fast_dot
from ...utils.common import intersects
from ...utils import get_rng
from ...utils.init_utils import zeros, xavier_uniform


class CausalRec(Recommender):
    """CausalRec: Causal Inference for Visual Debiasing in Visually-Aware Recommendation

    Parameters
    ----------
    k: int, optional, default: 10
        The dimension of the gamma latent factors.

    k2: int, optional, default: 10
        The dimension of the theta latent factors.

    n_epochs: int, optional, default: 20
        Maximum number of epochs for SGD.

    batch_size: int, optional, default: 100
        The batch size for SGD.

    learning_rate: float, optional, default: 0.001
        The learning rate for SGD.

    lambda_w: float, optional, default: 0.01
        The regularization hyper-parameter for latent factor weights.

    lambda_b: float, optional, default: 0.01
        The regularization hyper-parameter for biases.

    lambda_e: float, optional, default: 0.0
        The regularization hyper-parameter for embedding matrix E and beta prime vector.
    
    mean_feat: torch.tensor, required, default: None
        The mean feature of all item embeddings serving as the no-treatment during causal inference.
    
    tanh: int, optional, default: 0
        The number of tanh layers on the visual feature transformation.
    
    lambda_2: float, optional, default: 0.8
        The coefficient controlling the elimination of the visual bias in Eq. (28).

    use_gpu: boolean, optional, default: True
        Whether or not to use GPU to speed up training.

    trainable: boolean, optional, default: True
        When False, the model is not trained and Cornac assumes that the model already \
        pre-trained (U and V are not None).

    verbose: boolean, optional, default: True
        When True, running logs are displayed.

    init_params: dictionary, optional, default: None
        Initial parameters, e.g., init_params = {'Bi': beta_item, 'Gu': gamma_user,
        'Gi': gamma_item, 'Tu': theta_user, 'E': emb_matrix, 'Bp': beta_prime}

    seed: int, optional, default: None
        Random seed for weight initialization.

    References
    ----------
    * Qiu R., Wang S., Chen Z., Yin H., Huang Z. (2021). CausalRec: Causal Inference for Visual Debiasing in Visually-Aware Recommendation.
    """
    
    def __init__(
            self,
            name="CausalRec",
            k=10,
            k2=10,
            n_epochs=50,
            batch_size=100,
            learning_rate=0.005,
            lambda_w=0.01,
            lambda_b=0.01,
            lambda_e=0.0,
            mean_feat=None,
            tanh=0,
            lambda_2=0.8,
            use_gpu=False,
            trainable=True,
            verbose=True,
            init_params=None,
            seed=None,
    ):
        super().__init__(name=name, trainable=trainable, verbose=verbose)
        self.k = k
        self.k2 = k2
        self.n_epochs = n_epochs
        self.batch_size = batch_size
        self.learning_rate = learning_rate
        self.lambda_w = lambda_w
        self.lambda_b = lambda_b
        self.lambda_e = lambda_e
        self.mean_feat = mean_feat
        self.tanh = tanh
        self.lambda_2 = lambda_2
        self.use_gpu = use_gpu
        self.seed = seed
        
        # Init params if provided
        self.init_params = {} if init_params is None else init_params
        self.beta_item = self.init_params.get("Bi", None)
        self.gamma_user = self.init_params.get("Gu", None)
        self.gamma_item = self.init_params.get("Gi", None)
        self.theta_user = self.init_params.get("Tu", None)
        self.emb_matrix = self.init_params.get("E", None)
        self.emb_ind_matrix = self.init_params.get("E_ind", None)
        if self.tanh == 2:
            self.emb_matrix2 = self.init_params.get("E2", None)
            self.emb_ind_matrix2 = self.init_params.get("E_ind2", None)
        self.beta_prime = self.init_params.get("Bp", None)
    
    def _init(self, n_users, n_items, features):
        rng = get_rng(self.seed)
        
        self.beta_item = zeros(n_items) if self.beta_item is None else self.beta_item
        if self.gamma_user is None:
            self.gamma_user = xavier_uniform((n_users, self.k), rng)
        if self.gamma_item is None:
            self.gamma_item = xavier_uniform((n_items, self.k), rng)
        if self.theta_user is None:
            self.theta_user = xavier_uniform((n_users, self.k2), rng)
        if self.emb_matrix is None:
            self.emb_matrix = xavier_uniform((features.shape[1], self.k2), rng)
        if self.emb_ind_matrix is None:
            self.emb_ind_matrix = xavier_uniform((features.shape[1], self.k), rng)
        if self.tanh == 2:
            if self.emb_matrix2 is None:
                self.emb_matrix2 = xavier_uniform((self.k2, self.k2), rng)
            if self.emb_ind_matrix2 is None:
                self.emb_ind_matrix2 = xavier_uniform((self.k, self.k), rng)
        if self.beta_prime is None:
            self.beta_prime = xavier_uniform((features.shape[1], 1), rng)
        
        # pre-computed for faster evaluation
        self.theta_item = np.matmul(features, self.emb_matrix)
        self.visual_bias = np.matmul(features, self.beta_prime).ravel()
        self.direct_theta_item = np.matmul(features, self.emb_ind_matrix)
    
    def fit(self, train_set, val_set=None):
        """Fit the model to observations.

        Parameters
        ----------
        train_set: :obj:`cornac.data.Dataset`, required
            User-Item preference data as well as additional modalities.

        val_set: :obj:`cornac.data.Dataset`, optional, default: None
            User-Item preference data for model selection purposes (e.g., early stopping).

        Returns
        -------
        self : object
        """
        Recommender.fit(self, train_set, val_set)
        
        if train_set.item_image is None:
            raise CornacException("item_image modality is required but None.")
        
        # Item visual feature from CNN
        train_features = train_set.item_image.features[: self.train_set.total_items]
        train_features = train_features.astype(np.float32)
        self._init(
            n_users=train_set.total_users,
            n_items=train_set.total_items,
            features=train_features,
        )
        
        if self.trainable:
            self._fit_torch(train_features)
        
        return self
    
    def _fit_torch(self, train_features):
        import torch
        
        def _l2_loss(*tensors):
            l2_loss = 0
            for tensor in tensors:
                l2_loss += tensor.pow(2).sum()
            return l2_loss / 2
        
        def _inner(a, b):
            return (a * b).sum(dim=1)
        
        dtype = torch.float
        device = (
            torch.device("cuda:0")
            if (self.use_gpu and torch.cuda.is_available())
            else torch.device("cpu")
        )
        
        F = torch.tensor(train_features, device=device, dtype=dtype)
        # Learned parameters
        Bi = torch.tensor(
            self.beta_item, device=device, dtype=dtype, requires_grad=True
        )
        Gu = torch.tensor(
            self.gamma_user, device=device, dtype=dtype, requires_grad=True
        )
        Gi = torch.tensor(
            self.gamma_item, device=device, dtype=dtype, requires_grad=True
        )
        Tu = torch.tensor(
            self.theta_user, device=device, dtype=dtype, requires_grad=True
        )
        E = torch.tensor(
            self.emb_matrix, device=device, dtype=dtype, requires_grad=True
        )
        Bp = torch.tensor(
            self.beta_prime, device=device, dtype=dtype, requires_grad=True
        )
        E_ind = torch.tensor(
            self.emb_ind_matrix, device=device, dtype=dtype, requires_grad=True
        )
        mean_feat = torch.tensor(
            [self.mean_feat], device=device, dtype=dtype, requires_grad=False
        )
        param = [Bi, Gu, Gi, Tu, E, Bp, E_ind]
        
        if self.tanh == 2:
            E2 = torch.tensor(
                self.emb_matrix2, device=device, dtype=dtype, requires_grad=True
            )
            E_ind2 = torch.tensor(
                self.emb_ind_matrix2, device=device, dtype=dtype, requires_grad=True
            )
            param.append(E2)
            param.append(E_ind2)
        
        optimizer = torch.optim.Adam(param, lr=self.learning_rate)
        
        for epoch in range(1, self.n_epochs + 1):
            sum_loss = 0.0
            count = 0
            progress_bar = tqdm(
                total=self.train_set.num_batches(self.batch_size),
                desc="Epoch {}/{}".format(epoch, self.n_epochs),
                disable=not self.verbose,
            )
            for batch_u, batch_i, batch_j in self.train_set.uij_iter(
                    self.batch_size, shuffle=True
            ):
                gamma_u = Gu[batch_u]
                theta_u = Tu[batch_u]
                
                beta_i = Bi[batch_i]
                beta_j = Bi[batch_j]
                gamma_i = Gi[batch_i]
                gamma_j = Gi[batch_j]
                feat_i = F[batch_i]
                feat_j = F[batch_j]
                
                if self.tanh == 0:
                    direct_feat_i = feat_i.mm(E)
                    ind_feat_i = feat_i.mm(E_ind)
                elif self.tanh == 1:
                    direct_feat_i = torch.tanh(feat_i.mm(E))
                    ind_feat_i = torch.tanh(feat_i.mm(E_ind))
                elif self.tanh == 2:
                    direct_feat_i = torch.tanh(torch.tanh(feat_i.mm(E)).mm(E2))
                    ind_feat_i = torch.tanh(torch.tanh(feat_i.mm(E_ind)).mm(E_ind2))
                
                i_m = beta_i + _inner(gamma_u, gamma_i) + _inner(gamma_u, gamma_i * ind_feat_i)
                i_n = _inner(theta_u, direct_feat_i) + feat_i.mm(Bp)
                
                if self.tanh == 0:
                    direct_feat_j = feat_j.mm(E)
                    ind_feat_j = feat_j.mm(E_ind)
                elif self.tanh == 1:
                    direct_feat_j = torch.tanh(feat_j.mm(E))
                    ind_feat_j = torch.tanh(feat_j.mm(E_ind))
                elif self.tanh == 2:
                    direct_feat_j = torch.tanh(torch.tanh(feat_j.mm(E)).mm(E2))
                    ind_feat_j = torch.tanh(torch.tanh(feat_j.mm(E_ind)).mm(E_ind2))
                
                j_m = beta_j + _inner(gamma_u, gamma_j) + _inner(gamma_u, gamma_j * ind_feat_j)
                j_n = _inner(theta_u, direct_feat_j) + feat_j.mm(Bp)
                
                i_score = torch.sigmoid(i_m + i_n) * torch.sigmoid(i_m) * torch.sigmoid(i_n)
                j_score = torch.sigmoid(j_m + j_n) * torch.sigmoid(j_m) * torch.sigmoid(j_n)
                
                log_likelihood = torch.nn.functional.logsigmoid(i_score - j_score).sum()
                log_likelihood_m = torch.nn.functional.logsigmoid(i_m - j_m).sum()
                log_likelihood_n = torch.nn.functional.logsigmoid(i_n - j_n).sum()
                
                if self.tanh < 2:
                    l2_e = _l2_loss(E, Bp, E_ind)
                else:
                    l2_e = _l2_loss(E, Bp, E_ind, E2, E_ind2)
                
                reg = (
                        _l2_loss(gamma_u, gamma_i, gamma_j, theta_u) * self.lambda_w
                        + _l2_loss(beta_i) * self.lambda_b
                        + _l2_loss(beta_j) * self.lambda_b / 10
                        + l2_e * self.lambda_e
                )
                
                loss = -log_likelihood + reg - log_likelihood_m - log_likelihood_n
                
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
                
                sum_loss += loss.data.item()
                count += len(batch_u)
                if count % (self.batch_size * 10) == 0:
                    progress_bar.set_postfix(loss=(sum_loss / count))
                progress_bar.update(1)
            progress_bar.close()
        
        print("Optimization finished!")
        
        self.beta_item = Bi.data.cpu().numpy()
        self.gamma_user = Gu.data.cpu().numpy()
        self.gamma_item = Gi.data.cpu().numpy()
        self.theta_user = Tu.data.cpu().numpy()
        self.emb_matrix = E.data.cpu().numpy()
        self.beta_prime = Bp.data.cpu().numpy()
        self.emb_dir_matrix = E_ind.data.cpu().numpy()
        # pre-computed for faster evaluation
        self.theta_item = F.mm(E)
        if self.tanh == 0:
            self.theta_item = self.theta_item.data.cpu().numpy()
        elif self.tanh == 1:
            self.theta_item = torch.tanh(self.theta_item).data.cpu().numpy()
        elif self.tanh == 2:
            self.theta_item = torch.tanh(torch.tanh(self.theta_item).mm(E2)).data.cpu().numpy()

        self.visual_bias = F.mm(Bp).squeeze().data.cpu().numpy()

        self.ind_theta_item = F.mm(E_ind)
        if self.tanh == 0:
            self.ind_theta_item = self.ind_theta_item.data.cpu().numpy()
        elif self.tanh == 1:
            self.ind_theta_item = torch.tanh(self.ind_theta_item).data.cpu().numpy()
        elif self.tanh == 2:
            self.ind_theta_item = torch.tanh(torch.tanh(self.ind_theta_item).mm(E_ind2)).data.cpu().numpy()

        self.beta_item_mean = Bi.mean().unsqueeze(dim=0).data.cpu().numpy()
        self.gamma_item_mean = Gi.mean(dim=0).unsqueeze(dim=0).data.cpu().numpy()

        self.mean_feat = mean_feat.mm(E_ind)
        if self.tanh == 0:
            self.mean_feat = self.mean_feat.data.cpu().numpy()
        elif self.tanh == 1:
            self.mean_feat = torch.tanh(self.mean_feat).data.cpu().numpy()
        elif self.tanh == 2:
            self.mean_feat = torch.tanh(torch.tanh(self.mean_feat).mm(E_ind2)).data.cpu().numpy()

    def score(self, user_idx, item_idx=None):
        """Predict the debiased scores/ratings of a user for an item.

        Parameters
        ----------
        user_idx: int, required
            The index of the user for whom to perform score prediction.

        item_idx: int, optional, default: None
            The index of the item for which to perform score prediction.
            If None, scores for all known items will be returned.
        
        Returns
        -------
        res : A scalar or a Numpy array
            Relative scores that the user gives to the item or to all known items

        """
        if item_idx is None:
            m_score = self.beta_item
            fast_dot(self.gamma_user[user_idx], self.gamma_item, m_score)
            fast_dot(self.gamma_user[user_idx], self.gamma_item * self.ind_theta_item, m_score)

            m_star = self.beta_item_mean
            fast_dot(self.gamma_user[user_idx], self.gamma_item_mean, m_star)
            fast_dot(self.gamma_user[user_idx], self.gamma_item_mean * self.mean_feat, m_star)
            
            n_score = self.visual_bias
            fast_dot(self.theta_user[user_idx], self.theta_item, n_score)

            return expit(m_score + n_score) * expit(m_score) * expit(n_score)\
                   - self.lambda_2 * expit(m_star + n_score) * expit(m_star) * expit(n_score)
        else:
            raise NotImplementedError("The sampled evaluation is not implemented!")



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================


from .rating import RatingMetric
from .rating import MAE
from .rating import RMSE
from .rating import MSE

from .ranking import RankingMetric
from .ranking import NDCG
from .ranking import NCRR
from .ranking import MRR
from .ranking import Precision
from .ranking import Recall
from .ranking import FMeasure
from .ranking import AUC
from .ranking import MAP



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import numpy as np
import scipy.sparse as sp

from ..recommender import Recommender
from ...exception import ScoreException


class SKMeans(Recommender):
    """Spherical k-means based recommender.

    Parameters
    ----------
    k: int, optional, default: 5
        The number of clusters.

    max_iter: int, optional, default: 100
        Maximum number of iterations.

    name: string, optional, default: 'Skmeans'
        The name of the recommender model.

    trainable: boolean, optional, default: True
        When False, the model is not trained and Cornac assumes that the model is already \
        trained. 
        
    tol : float, optional, default: 1e-6
        Relative tolerance with regards to skmeans' criterion to declare convergence.
        
    verbose: boolean, optional, default: False
        When True, some running logs are displayed.

    seed: int, optional, default: None
        Random seed for parameters initialization.

    init_par: numpy 1d array, optional, default: None
        The initial object parition, 1d array contaning the cluster label (int type starting from 0) \
        of each object (user). If par = None, then skmeans is initialized randomly.
      
    centroids: csc_matrix, shape (k,n_users)
        The maxtrix of cluster centroids.

    References
    ----------
    * Salah, Aghiles, Nicoleta Rogovschi, and Mohamed Nadif. "A dynamic collaborative filtering system \
    via a weighted clustering approach." Neurocomputing 175 (2016): 206-215.
    """

    def __init__(
            self,
            k=5,
            max_iter=100,
            name="Skmeans",
            trainable=True,
            tol=1e-6,
            verbose=True,
            seed=None,
            init_par=None,
    ):
        Recommender.__init__(self, name=name, trainable=trainable, verbose=verbose)
        self.k = k
        self.max_iter = max_iter
        self.tol = tol
        self.verbose = verbose
        self.seed = seed
        self.init_par = init_par
        self.centroids = None  # matrix of cluster centroids

    def fit(self, train_set, val_set=None):
        """Fit the model to observations.

        Parameters
        ----------
        train_set: :obj:`cornac.data.Dataset`, required
            User-Item preference data as well as additional modalities.

        val_set: :obj:`cornac.data.Dataset`, optional, default: None
            User-Item preference data for model selection purposes (e.g., early stopping).

        Returns
        -------
        self : object
        """
        Recommender.fit(self, train_set, val_set)

        X = self.train_set.matrix
        X = sp.csr_matrix(X)

        # Skmeans requires rows of X to have a unit L2 norm. We therefore need to make a copy of X as we should not modify the latter.
        X1 = X.copy()
        X1 = X1.multiply(
            sp.csc_matrix(1.0 / (np.sqrt(X1.multiply(X1).sum(1).A1) + 1e-20)).T
        )

        if self.trainable:
            from .skmeans import skmeans

            res = skmeans(
                X1,
                k=self.k,
                max_iter=self.max_iter,
                tol=self.tol,
                verbose=self.verbose,
                seed=self.seed,
                init_par=getattr(self, "final_par", self.init_par),
            )
            self.centroids = res["centroids"]
            self.final_par = res["partition"]
        else:
            print("%s is trained already (trainable = False)" % (self.name))

        self.user_center_sim = (
                X1 * self.centroids.T
        )  # user-centroid cosine similarity matrix
        del X1

        return self

    def score(self, user_idx, item_idx=None):
        """Predict the scores/ratings of a user for an item.

        Parameters
        ----------
        user_idx: int, required
            The index of the user for whom to perform score prediction.

        item_idx: int, optional, default: None
            The index of the item for which to perform score prediction.
            If None, scores for all known items will be returned.

        Returns
        -------
        res : A scalar or a Numpy array
            Relative scores that the user gives to the item or to all known items

        """
        if item_idx is None:
            if self.train_set.is_unk_user(user_idx):
                raise ScoreException(
                    "Can't make score prediction for (user_id=%d)" % user_idx
                )

            known_item_scores = self.centroids.multiply(
                self.user_center_sim[user_idx, :].T
            )
            known_item_scores = known_item_scores.sum(0).A1 / (
                    self.user_center_sim[user_idx, :].sum() + 1e-20
            )  # weighted average of cluster centroids
            return known_item_scores
        else:
            if self.train_set.is_unk_user(user_idx) or self.train_set.is_unk_item(
                    item_idx
            ):
                raise ScoreException(
                    "Can't make score prediction for (user_id=%d, item_id=%d)"
                    % (user_idx, item_idx)
                )

            user_pred = self.centroids[item_idx, :].multiply(
                self.user_center_sim[user_idx, :].T
            )
            # transform user_pred to a flatten array
            user_pred = user_pred.sum(0).A1 / (
                    self.user_center_sim[user_idx, :].sum() + 1e-20
            )  # weighted average of cluster centroids

            return user_pred



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import numpy as np
import scipy.optimize as opt


class Model:
    def __init__(
        self,
        n_user,
        n_item,
        n_vocab,
        alpha,
        beta_u,
        beta_i,
        gamma_u,
        gamma_i,
        k=10,
        lambda_text=0.1,
        l2_reg=0.001,
        max_iter=50,
        grad_iter=50,
    ):
        self.k = k
        self.lambda_text = lambda_text
        self.l2_reg = l2_reg
        self.grad_iter = grad_iter
        self.max_iter = max_iter
        self.n_item = n_item
        self.n_user = n_user
        self.n_vocab = n_vocab

        # Model parameters
        self.alpha = alpha
        self.beta_u = beta_u
        self.beta_i = beta_i
        self.gamma_u = gamma_u
        self.gamma_i = gamma_i

        self._init_params()

    def _init_params(self):
        params_length = np.array(
            [
                1,
                1,
                self.n_user,
                self.n_item,
                self.n_user * self.k,
                self.n_item * self.k,
                self.n_vocab * self.k,
            ]
        )
        self.params_idx = params_length.cumsum()

        self.params = np.zeros(params_length.sum())
        self.params[0] = self.alpha
        self.params[1] = 1.0  # kappa init
        self.params[self.params_idx[3] : self.params_idx[4]] = self.gamma_u.ravel()
        self.params[self.params_idx[4] : self.params_idx[5]] = self.gamma_i.ravel()

    def init_count(self, docs):
        # Counter
        self.item_topic_cnt = np.zeros(
            shape=(self.n_item, self.k), dtype=int
        )  # given item, count number of time each topic occur

        self.word_topic_cnt = np.zeros(
            shape=(self.n_vocab, self.k), dtype=int
        )  # given word, count number of time each topic occur

        self.item_word = np.zeros(
            shape=(self.n_item, 1), dtype=int
        )  # number of word for each item
        self.topic_cnt = np.zeros(
            shape=(1, self.k), dtype=int
        )  # number of time topic occur
        self.total_word = 0  # total number word in corpus
        self.background_word = np.zeros(
            shape=(self.n_vocab, 1), dtype=float
        )  # background weights for each word [n_vocab, 1]

        self.topic_assignment = list()

        for di in range(len(docs)):
            doc = docs[di]
            doc_len = len(doc)
            topics = np.random.randint(self.k, size=doc_len)
            self.topic_assignment.append(topics)
            self.item_word[di] = doc_len
            self.total_word += doc_len
            for wi in range(doc_len):
                topic = topics[wi]
                word = doc[wi]
                self.word_topic_cnt[word, topic] += 1
                self.item_topic_cnt[di, topic] += 1
                self.background_word[word] += 1
                self.topic_cnt[0, topic] += 1

        self.background_word /= self.total_word

    @staticmethod
    def _sampling_from_dist(prob):
        thr = prob.sum() * np.random.rand()
        new_topic = 0
        tmp = prob[new_topic]
        while tmp < thr:
            new_topic += 1
            tmp += prob[new_topic]
        return new_topic

    def assign_word_topics(self, docs):
        _, self.kappa, _, _, _, self.gamma_i, self.topic_word = self._get_view(
            self.params
        )

        for di in range(len(docs)):
            doc = docs[di]
            doc_len = len(doc)
            topics = self.topic_assignment[di]

            for wi in range(doc_len):
                old_topic = topics[wi]
                word = doc[wi]
                topic_score = np.exp(
                    self.kappa * self.gamma_i[di]
                    + self.background_word[word]
                    + self.topic_word[word]
                )
                topic_score = topic_score / np.sum(topic_score)
                new_topic = self._sampling_from_dist(topic_score)
                if new_topic != old_topic:
                    self.word_topic_cnt[word, old_topic] -= 1
                    self.word_topic_cnt[word, new_topic] += 1
                    self.topic_cnt[0, old_topic] -= 1
                    self.topic_cnt[0, new_topic] += 1
                    self.item_topic_cnt[di, old_topic] -= 1
                    self.item_topic_cnt[di, new_topic] += 1
                    self.topic_assignment[di][wi] = new_topic

        average = self.topic_word.sum(axis=1)[:, np.newaxis] / self.k
        self.topic_word -= average
        self.background_word += average

    def update_params(self, rating_data):
        res = opt.fmin_l_bfgs_b(
            self._func, x0=self.params, args=rating_data, maxiter=self.grad_iter
        )
        self.params = res[0]
        return res[1]

    def get_parameter(self):
        alpha, _, beta_u, beta_i, gamma_u, gamma_i, _ = self._get_view(self.params)
        return alpha.item(), beta_u, beta_i, gamma_u, gamma_i

    def _get_view(self, params):
        idx = self.params_idx

        alpha = params[0 : idx[0],]
        kappa = params[idx[0] : idx[1],]
        beta_u = params[idx[1] : idx[2],]
        beta_i = params[idx[2] : idx[3],]
        gamma_u = params[idx[3] : idx[4],].reshape(self.n_user, self.k)
        gamma_i = params[idx[4] : idx[5],].reshape(self.n_item, self.k)
        topic_word = params[idx[5] :,].reshape(self.n_vocab, self.k)

        return alpha, kappa, beta_u, beta_i, gamma_u, gamma_i, topic_word

    def _func(self, X, *args):
        user_data = args[0]
        item_data = args[1]
        R_user = user_data[1]
        R_item = item_data[1]
        grad = np.zeros_like(X)
        alpha, kappa, beta_u, beta_i, gamma_u, gamma_i, topic_word = self._get_view(
            params=X
        )
        dalpha, dkappa, dbeta_u, dbeta_i, dgamma_u, dgamma_i, dtopic_word = self._get_view(
            params=grad
        )
        cf_loss = 0.0
        reg_loss = 0.0
        corpus_likelihood = 0.0

        for i in range(self.n_user):
            idx_item = user_data[0][i]
            if not len(idx_item):  # user without any rating
                continue
            gamma_items = gamma_i[idx_item]
            R_i = R_user[i]
            pred = alpha + beta_u[i] + beta_i[idx_item] + gamma_items.dot(gamma_u[i])
            err = (pred - R_i).reshape(-1, 1)
            cf_loss += np.sum(err ** 2)

            total_err = np.sum(err)
            dalpha += 2 * total_err
            dbeta_u[i] += 2 * total_err
            dgamma_u[i] += 2 * np.sum(err * gamma_items, axis=0)

        for j in range(self.n_item):
            idx_user = item_data[0][j]
            if not len(idx_user):  # item without any rating
                continue
            gamma_users = gamma_u[idx_user]
            R_j = R_item[j]
            pred = alpha + beta_u[idx_user] + beta_i[j] + gamma_users.dot(gamma_i[j])
            err = (pred - R_j).reshape(-1, 1)
            total_err = np.sum(err)
            dbeta_i[j] += 2 * total_err
            dgamma_i[j] += 2 * np.sum(err * gamma_users, axis=0)

        if self.l2_reg > 0:
            reg_loss += self.l2_reg * np.sum(gamma_u ** 2)
            dgamma_u += 2 * self.l2_reg * gamma_u
            reg_loss += self.l2_reg * np.sum(gamma_i ** 2)
            dgamma_i += 2 * self.l2_reg * gamma_i

        e_theta = np.exp(self.kappa * self.gamma_i)
        t_z = e_theta.sum(axis=1, keepdims=True)
        corpus_likelihood += self.lambda_text * np.sum(
            self.item_topic_cnt * (self.kappa * self.gamma_i - np.log(t_z))
        )

        e_phi = np.exp(self.background_word + topic_word)
        word_z = e_phi.sum(axis=0, keepdims=True)
        corpus_likelihood += self.lambda_text * np.sum(
            self.word_topic_cnt * (self.background_word + topic_word - np.log(word_z))
        )

        q = -self.lambda_text * (self.item_topic_cnt - self.item_word * e_theta / t_z)
        dgamma_i += kappa * q
        dkappa += np.sum(gamma_i * q)
        dtopic_word += -self.lambda_text * (
            self.word_topic_cnt - self.topic_cnt * e_phi / word_z
        )

        loss = cf_loss + reg_loss - corpus_likelihood

        return loss, grad



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import tensorflow.compat.v1 as tf

from ..cdl.cdl import act_functions


class Model:
    def __init__(
        self,
        n_users,
        n_items,
        input_dim,
        seed,
        n_z,
        layers,
        loss_type,
        act_fn,
        U,
        V,
        lambda_u,
        lambda_v,
        lambda_r,
        lambda_w,
        lr=0.001,
    ):
        self.n_users = n_users
        self.n_items = n_items
        self.input_dim = input_dim
        self.n_z = n_z
        self.layers = layers
        self.loss_type = loss_type
        self.act_fn = act_fn
        self.seed = seed
        self.lambda_u = lambda_u
        self.lambda_v = lambda_v
        self.lambda_r = lambda_r
        self.lambda_w = lambda_w
        self.lr = lr
        self.U_init = tf.constant(U)
        self.V_init = tf.constant(V)

        self._build_graph()

    def _build_graph(self):
        self.x = tf.placeholder(tf.float32, [None, self.input_dim], name="text_input")
        self.ratings = tf.placeholder(
            dtype=tf.float32, shape=[self.n_users, None], name="rating_input"
        )
        self.C = tf.placeholder(
            dtype=tf.float32, shape=[self.n_users, None], name="C_input"
        )
        self.item_ids = tf.placeholder(dtype=tf.int32, name="item_input")

        x_recon = self._vae(self.x)

        with tf.variable_scope("cf_variable"):
            self.U = tf.get_variable(
                name="U", dtype=tf.float32, initializer=self.U_init
            )
            self.V = tf.get_variable(
                name="V", dtype=tf.float32, initializer=self.V_init
            )

        real_batch_size = tf.cast(tf.shape(self.item_ids)[0], tf.int32)
        V_batch = tf.reshape(
            tf.gather(self.V, self.item_ids), shape=[real_batch_size, self.n_z]
        )

        predictions = tf.matmul(self.U, V_batch, transpose_b=True)
        squared_error = tf.square(self.ratings - predictions)

        # CF loss
        rating_loss = tf.reduce_mean(
            tf.reduce_sum(tf.multiply(self.C, squared_error), 0)
        )
        v_loss = (
            self.lambda_v
            / self.lambda_r
            * tf.reduce_mean(tf.reduce_sum(tf.square(V_batch - self.z), 1))
        )
        self.cf_loss = rating_loss + v_loss + self.lambda_u * tf.nn.l2_loss(self.U)

        # VAE loss
        if self.loss_type == "rmse":
            gen_loss = tf.reduce_mean(tf.square(tf.subtract(self.x, x_recon)))
        elif self.loss_type == "cross-entropy":
            x_recon = tf.nn.sigmoid(x_recon, name="x_recon")
            gen_loss = -tf.reduce_mean(
                tf.reduce_sum(
                    self.x * tf.log(tf.maximum(x_recon, 1e-10))
                    + (1 - self.x) * tf.log(tf.maximum(1 - x_recon, 1e-10)),
                    1,
                )
            )
        else:
            raise ValueError("Invalid loss type {}".format(self.loss_type))

        latent_loss = 0.5 * tf.reduce_mean(
            tf.reduce_sum(
                tf.square(self.z_mean)
                + tf.exp(self.z_log_sigma_sq)
                - self.z_log_sigma_sq
                - 1,
                1,
            )
        )
        self.vae_loss = gen_loss + latent_loss + self.lambda_w * self.reg_loss

        cf_op = tf.train.AdamOptimizer(self.lr, name="cf_op")
        vae_op = tf.train.AdamOptimizer(self.lr, name="vae_op")

        cf_var_list, vae_var_list = [], []

        for var in tf.trainable_variables():
            if "cf_variable" in var.name:
                cf_var_list.append(var)
            elif "vae" in var.name:
                vae_var_list.append(var)

        gvs = cf_op.compute_gradients(self.cf_loss, var_list=cf_var_list)
        capped_gvs = [(tf.clip_by_value(grad, -5.0, 5.0), var) for grad, var in gvs]
        self.cf_update = cf_op.apply_gradients(capped_gvs)

        gvs = vae_op.compute_gradients(self.vae_loss, var_list=vae_var_list)
        capped_gvs = [(tf.clip_by_value(grad, -5.0, 5.0), var) for grad, var in gvs]
        self.vae_update = vae_op.apply_gradients(capped_gvs)

    def _vae(self, X):
        act_fn = act_functions.get(self.act_fn, None)
        if act_fn is None:
            raise ValueError(
                "Invalid type of activation function {}\n"
                "Supported functions: {}".format(act_fn, act_functions.keys())
            )

        with tf.variable_scope("vae/inference"):
            rec = {
                "W1": tf.get_variable(
                    "W1",
                    [self.input_dim, self.layers[0]],
                    initializer=tf.keras.initializers.glorot_uniform(seed=self.seed),
                    dtype=tf.float32,
                ),
                "b1": tf.get_variable(
                    "b1",
                    [self.layers[0]],
                    initializer=tf.constant_initializer(0.0),
                    dtype=tf.float32,
                ),
                "W2": tf.get_variable(
                    "W2",
                    [self.layers[0], self.layers[1]],
                    initializer=tf.keras.initializers.glorot_uniform(seed=self.seed),
                    dtype=tf.float32,
                ),
                "b2": tf.get_variable(
                    "b2",
                    [self.layers[1]],
                    initializer=tf.constant_initializer(0.0),
                    dtype=tf.float32,
                ),
                "W_z_mean": tf.get_variable(
                    "W_z_mean",
                    [self.layers[1], self.n_z],
                    initializer=tf.keras.initializers.glorot_uniform(seed=self.seed),
                    dtype=tf.float32,
                ),
                "b_z_mean": tf.get_variable(
                    "b_z_mean",
                    [self.n_z],
                    initializer=tf.constant_initializer(0.0),
                    dtype=tf.float32,
                ),
                "W_z_log_sigma": tf.get_variable(
                    "W_z_log_sigma",
                    [self.layers[1], self.n_z],
                    initializer=tf.keras.initializers.glorot_uniform(seed=self.seed),
                    dtype=tf.float32,
                ),
                "b_z_log_sigma": tf.get_variable(
                    "b_z_log_sigma",
                    [self.n_z],
                    initializer=tf.constant_initializer(0.0),
                    dtype=tf.float32,
                ),
            }

        h1 = act_fn(tf.matmul(X, rec["W1"]) + rec["b1"])
        h2 = act_fn(tf.matmul(h1, rec["W2"]) + rec["b2"])

        self.z_mean = tf.matmul(h2, rec["W_z_mean"]) + rec["b_z_mean"]
        self.z_log_sigma_sq = tf.matmul(h2, rec["W_z_log_sigma"]) + rec["b_z_log_sigma"]

        eps = tf.random_normal(
            shape=tf.shape(self.z_mean),
            mean=0,
            stddev=1,
            seed=self.seed,
            dtype=tf.float32,
        )

        self.z = (
            self.z_mean + tf.sqrt(tf.maximum(tf.exp(self.z_log_sigma_sq), 1e-10)) * eps
        )

        with tf.variable_scope("vae/generation"):
            gen = {
                "W2": tf.get_variable(
                    "W2",
                    [self.n_z, self.layers[1]],
                    initializer=tf.keras.initializers.glorot_uniform(seed=self.seed),
                    dtype=tf.float32,
                ),
                "b2": tf.get_variable(
                    "b2",
                    [self.layers[1]],
                    initializer=tf.constant_initializer(0.0),
                    dtype=tf.float32,
                ),
                "W1": tf.transpose(rec["W2"]),
                "b1": rec["b1"],
                "W_x": tf.transpose(rec["W1"]),
                "b_x": tf.get_variable(
                    "b_x",
                    [self.input_dim],
                    initializer=tf.constant_initializer(0.0),
                    dtype=tf.float32,
                ),
            }

        self.reg_loss = (
            tf.nn.l2_loss(rec["W1"])
            + tf.nn.l2_loss(rec["W2"])
            + tf.nn.l2_loss(gen["W1"])
            + tf.nn.l2_loss(gen["W_x"])
        )

        h2 = act_fn(tf.matmul(self.z, gen["W2"]) + gen["b2"])
        h1 = act_fn(tf.matmul(h2, gen["W1"]) + gen["b1"])
        x_recon = tf.matmul(h1, gen["W_x"]) + gen["b_x"]

        return x_recon
